"""
A very simple lock mechanism to prevent multiple downloaders being started for the same Moodle.

The functions are not resistant to high frequency calls.
Raise conditions will occur!
"""

from pathlib import Path


class LockError(Exception):
    """An Exception which gets thrown if a Downloader is already running."""

    pass


def lock(dir_path: str):
    """Test if a lock is already set in a directory, if not it creates the lock."""
    path = Path(dir_path) / 'running.lock'
    if Path(path).exists():
        raise LockError(f'A downloader is already running. Delete {str(path)} if you think this is wrong.')
    Path(path).touch()


def unlock(dir_path: str):
    """Remove a lock in a directory."""
    path = Path(dir_path) / 'running.lock'
    try:
        Path(path).unlink()
    except OSError:
        pass



import logging
import threading

from queue import Queue, Empty

from moodle_dl.state_recorder.state_recorder import StateRecorder


class Downloader(threading.Thread):
    """
    Downloader processes the queue and puts an
    URL target back into the queue if an error occurs.
    """

    def __init__(
        self,
        queue: Queue,
        report: [],
        state_recorder: StateRecorder,
        thread_id: int,
        db_lock: threading.Lock,
    ):
        """
        Initiates a downloader thread.
        """
        threading.Thread.__init__(self)
        self.daemon = True

        self.queue = queue
        self.report = report
        self.thread_id = thread_id
        self.db_lock = db_lock
        self.state_recorder = state_recorder

    def run(self):
        """
        Work the queue until it is empty.
        """
        logging.debug('T%s - Downloader thread was started', self.thread_id)
        while self.queue.empty() is False:
            try:
                # raise condition
                url_target = self.queue.get(False)
            except Empty:
                break

            response = url_target.download(self.thread_id)

            # All information is still saved in url_target

            # If a download fails, add it to the error report.
            if response is False:
                logging.debug('T%s - URLTarget reports failure!', self.thread_id)
                self.report['failure'].append(url_target)

            # If a download was successful, store it in the database.
            elif response is True:
                logging.debug('T%s - URLTarget reports success!', self.thread_id)
                self.db_lock.acquire()
                self.state_recorder.save_file(url_target.file, url_target.course.id, url_target.course.fullname)
                self.db_lock.release()
                self.report['success'].append(url_target)

            self.queue.task_done()

        logging.debug('T%s - Downloader thread is finished', self.thread_id)



import os
import logging

from pathlib import Path

from moodle_dl.utils.logger import Log
from moodle_dl.moodle_connector.request_helper import RequestHelper, RequestRejectedError


class CookieHandler:
    """
    Fetches and saves the cookies of Moodle.
    """

    def __init__(self, request_helper: RequestHelper, version: int, storage_path: str):
        self.request_helper = request_helper
        self.version = version
        self.storage_path = storage_path
        self.cookies_path = str(Path(storage_path) / 'Cookies.txt')

        self.moodle_test_url = self.request_helper.url_base

    def fetch_autologin_key(self, privatetoken: str) -> {str: str}:

        # do this only if version is greater then 3.2
        # because tool_mobile_get_autologin_key will fail
        if self.version < 2016120500:
            return None

        print('\rDownloading autologin key\033[K', end='')

        extra_data = {'privatetoken': privatetoken}

        try:
            autologin_key_result = self.request_helper.post_REST('tool_mobile_get_autologin_key', extra_data)
            return autologin_key_result
        except RequestRejectedError as e:
            logging.debug("Cookie lockout: %s", e)  # , extra={'exception': e}
            return None

    def test_cookies(self) -> bool:
        """Test if cookies are valide

        Returns:
            bool: True if valide
        """

        logging.debug('Testing cookies with this URL %s', self.moodle_test_url)

        response, dummy = self.request_helper.get_URL(self.moodle_test_url, self.cookies_path)

        response_text = response.text

        if response_text.find('login/logout.php') >= 0:
            return True
        return False

    def check_and_fetch_cookies(self, privatetoken: str, userid: str) -> bool:
        if os.path.exists(self.cookies_path):
            # test if still logged in.

            if self.test_cookies():
                logging.debug('Cookies are still valid')
                return True

            warning_msg = 'Moodle cookie has expired, an attempt is made to generate a new cookie.'
            logging.warning(warning_msg)
            Log.warning('\r' + warning_msg + '\033[K')

        if privatetoken is None:
            error_msg = (
                'Moodle Cookies are not retrieved because no private token is set.'
                + ' To set a private token, use the `--new-token` option (if necessary also with `--sso`).'
            )
            logging.warning(error_msg)
            Log.error('\r' + error_msg + '\033[K')
            return False

        autologin_key = self.fetch_autologin_key(privatetoken)

        if autologin_key is None:
            error_msg = 'Failed to download autologin key!'
            logging.debug(error_msg)
            print('')
            Log.error(error_msg)
            return False

        print('\rDownloading cookies\033[K', end='')

        post_data = {'key': autologin_key.get('key', ''), 'userid': userid}
        url = autologin_key.get('autologinurl', '')

        cookies_response, dummy = self.request_helper.post_URL(url, post_data, self.cookies_path)

        logging.debug('Autologin redirected to %s', cookies_response.url)

        if self.test_cookies():
            return True
        else:
            error_msg = 'Failed to generate cookies!'
            logging.debug(error_msg)
            print('')
            Log.error(error_msg)
            return False



# -*- coding: utf-8 -*-
# This file is based on this gist:
# http://code.activestate.com/recipes/134892/
# So real authors are DannyYoo and company.

import msvcrt


def readchar(blocking=False):
    """reads a single character from the input stream. Retruns None if none is avalable.
    If blocking=True the function waits for the next character."""

    if blocking or msvcrt.kbhit():
        # manual byte decoding because some bytes in windows are not utf-8 encodable.
        return chr(int.from_bytes(msvcrt.getch(), "big"))
    else:
        return None


def readkey():
    """Get a single character on Windows. If an extended key is pressed, the
    Windows scan code is translated into a the unicode sequences readchar
    expects (see key_windows.py)."""

    ch = readchar(blocking=True)

    if ch == "\x03":
        raise KeyboardInterrupt

    # if it is a normal character:
    if ch not in "\x00\xe0":
        return ch

    # if it is a scpeal key, read second half:
    ch2 = readchar()

    return "\x00" + ch2



# coding: utf-8
from __future__ import unicode_literals

import re
import html

from yt_dlp.extractor.common import InfoExtractor
from yt_dlp.extractor.kaltura import KalturaIE
from yt_dlp.utils import (
    ExtractorError,
    urlencode_postdata,
    extract_attributes,
)


class KalvidresIE(InfoExtractor):
    IE_NAME = 'kalvidres'
    _VALID_URL = r'(?P<scheme>https?://)(?P<host>[^/]+)(?P<path>.*)?/mod/kalvidres/view.php\?.*?id=(?P<id>\d+)'
    _LAUNCH_FORM = 'ltiLaunchForm'

    def _real_extract(self, url):
        mobj = re.match(self._VALID_URL, url)
        # scheme = mobj.group('scheme')
        # host = mobj.group('host')
        # path = mobj.group('path')
        video_id = mobj.group('id')

        # Extract launch URL
        view_webpage = self._download_webpage(url, video_id, 'Downloading kalvidres video view webpage')
        mobj = re.search(r'<iframe[^>]+class="kaltura-player-iframe"[^>]+src=(["\'])(?P<url>[^"\']+)\1', view_webpage)
        if not mobj:
            raise ExtractorError('Unable to extract kalvidres launch url')

        launch_url = html.unescape(mobj.group('url'))

        # Get launch parameters
        launch_webpage = self._download_webpage(launch_url, video_id, 'Downloading kalvidres launch webpage')
        launch_inputs = self._form_hidden_inputs(self._LAUNCH_FORM, launch_webpage)
        launch_form_str = self._search_regex(
            fr'(?P<form><form[^>]+?id=(["\']){self._LAUNCH_FORM}\2[^>]*>)', launch_webpage, 'login form', group='form'
        )

        action_url = extract_attributes(launch_form_str).get('action')

        # Launch kalvidres video app
        submit_page, dummy = self._download_webpage_handle(
            action_url, video_id, 'Launch kalvidres app', data=urlencode_postdata(launch_inputs)
        )

        mobj = re.search(r'window.location.href = \'(?P<url>[^\']+)\'', submit_page)
        if not mobj:
            raise ExtractorError('Unable to extract kalvidres redirect url')

        # Follow kalvidres video app redirect
        redirect_page, dummy = self._download_webpage_handle(
            html.unescape(mobj.group('url')), video_id, 'Follow kalvidres redirect'
        )

        kultura_url = KalturaIE._extract_url(redirect_page)  # pylint: disable=protected-access
        if not kultura_url:
            raise ExtractorError('Unable to extract kaltura url')

        return {
            '_type': 'url',
            'url': kultura_url,
            'ie_key': 'Kaltura',
        }



from moodle_dl.moodle_connector.request_helper import RequestHelper, RequestRejectedError
from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.path_tools import PathTools
from moodle_dl.moodle_connector.moodle_constants import moodle_html_footer, moodle_html_header


class QuizzesHandler:
    """
    Fetches and parses the various endpoints in Moodle for quiz entries.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_quizzes(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Quizzes List for all courses from the
        Moodle system
        @return: A Dictionary of all quizzes,
                 indexed by courses, then quizzes
        """
        # do this only if version is greater then 3.1
        # because mod_quiz_get_quizzes_by_courses will fail
        if self.version < 2016052300:
            return {}

        print('\rDownloading quizzes information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        quizzes_result = self.request_helper.post_REST('mod_quiz_get_quizzes_by_courses', extra_data)

        quizzes = quizzes_result.get('quizzes', [])

        result = {}
        for quiz in quizzes:
            # This is the instance id with which we can make the API queries.
            quiz_id = quiz.get('id', 0)
            quiz_name = quiz.get('name', 'unnamed quiz')
            quiz_intro = quiz.get('intro', '')
            quiz_course_module_id = quiz.get('coursemodule', 0)
            quiz_introfiles = quiz.get('introfiles', [])
            course_id = quiz.get('course', 0)

            # normalize
            for quiz_file in quiz_introfiles:
                file_type = quiz_file.get('type', '')
                if file_type is None or file_type == '':
                    quiz_file.update({'type': 'quiz_introfile'})

            if quiz_intro != '':
                # Add Intro File
                intro_file = {
                    'filename': 'Quiz intro',
                    'filepath': '/',
                    'description': quiz_intro,
                    'type': 'description',
                }
                quiz_introfiles.append(intro_file)

            quiz_entry = {
                quiz_course_module_id: {
                    'id': quiz_id,
                    'name': quiz_name,
                    'intro': quiz_intro,
                    'files': quiz_introfiles,
                }
            }

            course_dic = result.get(course_id, {})

            course_dic.update(quiz_entry)

            result.update({course_id: course_dic})

        return result

    def fetch_quizzes_files(self, userid: int, quizzes: {}) -> {}:
        """
        Fetches for the quizzes list of all courses the additionally
        entries. This is kind of waste of resources, because there
        is no API to get all entries at once.
        @param userid: the user id.
        @param quizzes: the dictionary of quizzes of all courses.
        @return: A Dictionary of all quizzes,
                 indexed by courses, then quizzes
        """
        # do this only if version is greater then 3.1
        # because mod_quiz_get_user_attempts will fail
        if self.version < 2016052300:
            return quizzes

        counter = 0
        total = 0
        # count total quizzes for nice console output
        for course_id in quizzes:
            for quiz_id in quizzes[course_id]:
                total += 1

        for course_id in quizzes:
            for quiz_id in quizzes[course_id]:
                counter += 1
                quiz = quizzes[course_id][quiz_id]
                real_id = quiz.get('id', 0)
                data = {'quizid': real_id, 'userid': userid, 'status': 'all'}

                shorted_quiz_name = quiz.get('name', '')
                if len(shorted_quiz_name) > 17:
                    shorted_quiz_name = shorted_quiz_name[:15] + '..'

                print(
                    (
                        '\r'
                        + 'Downloading quiz infos'
                        + f' {counter:3d}/{total:3d}'
                        + f' [{shorted_quiz_name:<17}|{course_id:6}]\033[K'
                    ),
                    end='',
                )

                attempts_result = self.request_helper.post_REST('mod_quiz_get_user_attempts', data)
                attempts = attempts_result.get('attempts', [])

                quiz_files = self._get_files_of_attempts(attempts, quiz.get('name', ''))
                quiz['files'] += quiz_files

        return quizzes

    def _get_files_of_attempts(self, attempts: [], quiz_name: str) -> []:
        result = []

        for attempt in attempts:
            attempt_id = attempt.get('id', 0)
            attempt_state = attempt.get('state', 'unknown')

            attempt_filename = PathTools.to_valid_name(
                quiz_name + ' (attempt ' + str(attempt_id) + ' ' + attempt_state + ')'
            )

            shorted_quiz_name = quiz_name
            if len(shorted_quiz_name) > 17:
                shorted_quiz_name = shorted_quiz_name[:15] + '..'

            data = {'attemptid': attempt_id}

            try:
                if attempt_state == 'finished':
                    attempt_result = self.request_helper.post_REST('mod_quiz_get_attempt_review', data)
                elif attempt_state == 'inprogress':
                    attempt_result = self.request_helper.post_REST('mod_quiz_get_attempt_summary', data)
                else:
                    continue
            except RequestRejectedError:
                continue

            questions = attempt_result.get('questions', [])

            # build quiz HTML
            quiz_html = moodle_html_header
            for question in questions:

                question_html = question.get('html', '').split('<script>')[0]
                if question_html is None:
                    question_html = ''

                quiz_html += question_html + '\n'

                question_files = question.get('responsefileareas', [])
                for question_file in question_files:
                    file_type = question_file.get('type', '')
                    if file_type is None or file_type == '':
                        question_file.update({'type': 'quiz_file'})
                    result.append(question_file)

            quiz_html += moodle_html_footer
            attempt_file = {
                'filename': attempt_filename,
                'filepath': '/',
                'timemodified': 0,
                'html': quiz_html,
                'type': 'html',
                'no_search_for_urls': True,
            }
            result.append(attempt_file)

        return result



from moodle_dl.state_recorder.course import Course
from moodle_dl.moodle_connector.request_helper import RequestHelper


class FoldersHandler:
    """
    Fetches and parses the various endpoints in Moodle.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_folders(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Folder List for all courses from the
        Moodle system
        @return: A Dictionary of all folders,
                 indexed by courses, then folders
        """
        # do this only if version is greater then 3.3
        # because mod_folder_get_folders_by_courses will fail
        if self.version < 2017051500:
            return {}

        print('\rDownloading folders information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        folders_result = self.request_helper.post_REST('mod_folder_get_folders_by_courses', extra_data)

        folders = folders_result.get('folders', [])

        result = {}
        for folder in folders:
            folder_id = folder.get('id', 0)
            folder_name = folder.get('name', 'unnamed folder')
            folder_intro = folder.get('intro', '')
            folder_course_module_id = folder.get('coursemodule', 0)
            folder_files = folder.get('introfiles', [])
            course_id = folder.get('course', 0)
            folder_timemodified = folder.get('timemodified', 0)

            # normalize
            for folder_file in folder_files:
                file_type = folder_file.get('type', '')
                if file_type is None or file_type == '':
                    folder_file.update({'type': 'folder_file'})

            if folder_intro != '':
                # Add intro file
                intro_file = {
                    'filename': 'Folder intro',
                    'filepath': '/',
                    'description': folder_intro,
                    'timemodified': folder_timemodified,
                    'filter_urls_during_search_containing': ['/mod_folder/intro'],
                    'type': 'description',
                }
                folder_files.append(intro_file)

            folder_entry = {
                folder_course_module_id: {
                    'id': folder_id,
                    'name': folder_name,
                    'intro': folder_intro,
                    'files': folder_files,
                }
            }

            course_dic = result.get(course_id, {})
            course_dic.update(folder_entry)
            result.update({course_id: course_dic})

        return result



import json

from moodle_dl.moodle_connector.request_helper import RequestHelper
from moodle_dl.state_recorder.course import Course


class FirstContactHandler:
    """
    Fetches and parses the various endpoints in Moodle.
    """

    def __init__(self, request_helper: RequestHelper):
        self.request_helper = request_helper
        # oldest supported Moodle version
        self.version = 2011120500

    def fetch_userid_and_version(self) -> str:
        """
        Ask the Moodle system for the user id.
        @return: the userid
        """
        result = self.request_helper.post_REST('core_webservice_get_site_info')

        if 'userid' not in result:
            raise RuntimeError('Error could not receive your user ID!')
        userid = result.get('userid', '')

        version = result.get('version', '2011120500')

        try:
            version = int(version.split('.')[0])
        except Exception as e:
            raise RuntimeError(f'Error could not parse version string: "{version}" Error: {e}')

        self.version = version
        return userid, version

    def fetch_courses(self, userid: str) -> [Course]:
        """
        Queries the Moodle system for all courses the user
        is enrolled in.
        @param userid: the user id
        @return: A list of courses
        """
        data = {'userid': userid}

        courses = self.request_helper.post_REST('core_enrol_get_users_courses', data)

        results = []
        for course in courses:
            results.append(Course(course.get('id', 0), course.get('fullname', '')))
            # We could also extract here the course summary and intro files
        return results

    def fetch_all_visible_courses(self, log_all_courses_to: str = None) -> [Course]:
        """
        Queries the Moodle system for all courses available on the system and returns:
        @return: A list of all visible courses
        """
        # API is only available since version 3.2
        if self.version < 2016120500:
            return []

        result = self.request_helper.post_REST('core_course_get_courses_by_field', timeout=1200)
        if log_all_courses_to is not None:
            with open(log_all_courses_to, 'w', encoding='utf-8') as log_file:
                log_file.write(json.dumps(result, indent=4, ensure_ascii=False))
        courses = result.get('courses', [])

        results = []
        for course in courses:
            if course.get('visible', 0) == 1:
                results.append(Course(course.get('id', 0), course.get('fullname', '')))
        return results

    def fetch_courses_info(self, course_ids: [int]) -> [Course]:
        """
        Queries the Moodle system for info about courses in a list.
        @param course_ids: A list of courses ids
        @return: A list of courses
        """
        # API is only available since version 3.2
        if len(course_ids) == 0 or self.version < 2016120500:
            return []

        data = {
            "field": "ids",
            "value": ",".join(list(map(str, course_ids))),
        }

        result = self.request_helper.post_REST('core_course_get_courses_by_field', data)
        courses = result.get('courses', [])

        results = []
        for course in courses:
            results.append(Course(course.get('id', 0), course.get('fullname', '')))
            # We could also extract here the course summary and intro files
        return results

    def fetch_sections(self, course_id: int) -> [{}]:
        """
        Fetches the Sections List for a course from the Moodle system
        @param course_id: The id of the requested course.
        @return: A List of all section dictionaries
        """
        data = {'courseid': course_id}
        if self.version >= 2015051100:
            data = {
                'courseid': course_id,
                'options[0][name]': 'excludemodules',
                'options[0][value]': 'true',
                'options[1][name]': 'excludecontents',
                'options[1][value]': 'true',
            }
        course_sections = self.request_helper.post_REST('core_course_get_contents', data)

        sections = []
        for section in course_sections:
            sections.append({"id": section.get("id"), "name": section.get("name")})

        return sections



import requests
import ssl
import urllib3
"""
A custom helper that fixes OpenSSL incompatibility issues.
See https://stackoverflow.com/a/71646353 for more details.
"""
class CustomHttpAdapter (requests.adapters.HTTPAdapter):
    '''Transport adapter that allows us to use custom ssl_context.'''

    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, ssl_context=self.ssl_context)


def configure_ssl_context(context):
    """
    Customize SSL context to circumvent issues where the client uses
    OpenSSL 3.X.X and the server only supports OpenSSL 1.X.X.
    """
    context.options |= 0x4 # set ssl.OP_LEGACY_SERVER_CONNECT bit (https://bugs.python.org/issue44888)


def custom_session(verify_cert):
    """
    Defines a new session with custom SSL context to support edge cases with OpenSSL 3.X.X
    """
    session = requests.Session()
    if verify_cert:
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    else:
        ctx = ssl._create_unverified_context()
    configure_ssl_context(ctx)
    session.mount('https://', CustomHttpAdapter(ctx))
    return session



from moodle_dl.state_recorder.course import Course
from moodle_dl.moodle_connector.request_helper import RequestHelper


class PagesHandler:
    """
    Fetches and parses the various endpoints in Moodle.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_pages(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Pages List for all courses from the
        Moodle system
        @return: A Dictionary of all pages,
                 indexed by courses, then pages
        """
        # do this only if version is greater then 3.3
        # because mod_page_get_pages_by_courses will fail
        if self.version < 2017051500:
            return {}

        print('\rDownloading pages information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        pages_result = self.request_helper.post_REST('mod_page_get_pages_by_courses', extra_data)

        pages = pages_result.get('pages', [])

        result = {}
        for page in pages:
            page_id = page.get('id', 0)
            page_name = page.get('name', 'unnamed page')
            page_intro = page.get('intro', '')
            page_content = page.get('content', '')
            page_course_module_id = page.get('coursemodule', 0)
            page_files = page.get('introfiles', [])
            page_files += page.get('contentfiles', [])
            course_id = page.get('course', 0)
            page_timemodified = page.get('timemodified', 0)

            # normalize
            for page_file in page_files:
                file_type = page_file.get('type', '')
                if file_type is None or file_type == '':
                    page_file.update({'type': 'page_file'})

            if page_intro != '':
                # Add intro file
                intro_file = {
                    'filename': 'Page intro',
                    'filepath': '/',
                    'description': page_intro,
                    'type': 'description',
                }
                page_files.append(intro_file)

            if page_content != '':
                # Add content file
                content_file = {
                    'filename': page_name,
                    'filepath': '/',
                    'html': page_content,
                    'filter_urls_during_search_containing': ['/mod_page/content/'],
                    'no_hash': True,
                    'type': 'html',
                    'timemodified': page_timemodified,
                    'filesize': len(page_content),
                }
                page_files.append(content_file)

            page_entry = {
                page_course_module_id: {
                    'id': page_id,
                    'name': page_name,
                    'intro': page_intro,
                    'files': page_files,
                }
            }

            course_dic = result.get(course_id, {})
            course_dic.update(page_entry)
            result.update({course_id: course_dic})

        return result



from yt_dlp.YoutubeDL import YoutubeDL

from moodle_dl.ydl_extractors.zoomus import ZoomUSIE
from moodle_dl.ydl_extractors.opencast import OpencastIE, OpencastPlaylistIE
from moodle_dl.ydl_extractors.helixmedia import HelixmediaIE
from moodle_dl.ydl_extractors.kalvidres import KalvidresIE
from moodle_dl.ydl_extractors.owncloud import OwncloudIE
from moodle_dl.ydl_extractors.opencast_lti import OpencastLTIIE
from moodle_dl.ydl_extractors.googledrive import GoogleDriveIE


def add_additional_extractors(ydl: YoutubeDL):
    additional_extractors = [
        OpencastIE(ydl),
        OpencastPlaylistIE(ydl),
        ZoomUSIE(ydl),
        HelixmediaIE(ydl),
        KalvidresIE(ydl),
        OwncloudIE(ydl),
        OpencastLTIIE(ydl),
        GoogleDriveIE(ydl),
    ]

    moodle_dl_ies = {}
    moodle_dl_ies_instances = {}
    for extractor in additional_extractors:
        ie_key = extractor.ie_key()
        moodle_dl_ies[ie_key] = extractor
        moodle_dl_ies_instances[ie_key] = extractor

    # We access protected member variables of the yt-dlp to add the extractors afterwards.
    # TODO: Use the new possibilities yt-dlp offers to add the extractors to yt-dlp.
    # pylint: disable=protected-access
    moodle_dl_ies.update(ydl._ies)
    moodle_dl_ies_instances.update(ydl._ies_instances)
    ydl._ies = moodle_dl_ies
    ydl._ies_instances = moodle_dl_ies_instances



from moodle_dl.state_recorder.file import File
from moodle_dl.download_service.path_tools import PathTools


class Course:
    def __init__(self, _id: int, fullname: str, files: [File] = None):
        self.id = _id
        self.fullname = PathTools.to_valid_name(fullname)
        if files is not None:
            self.files = files
        else:
            self.files = []

        self.overwrite_name_with = None
        self.create_directory_structure = True
        self.excluded_sections = []

    def __str__(self):
        message = 'Course ('

        message += f'id: {self.id}'
        message += f', fullname: "{self.fullname}"'
        message += f', overwrite_name_with: "{PathTools.to_valid_name(self.overwrite_name_with)}"'
        message += f', create_directory_structure: {self.create_directory_structure}'
        message += f', files: {len(self.files)}'
        message += ')'
        return message



import os
import asyncio
import certifi

import aioxmpp


class XmppShooter:

    """
    A basic XMPP shooter that will log in, send messages,
    and then log out.
    """

    @staticmethod
    def my_ssl_factory():
        ctx = aioxmpp.security_layer.default_ssl_context()
        ctx.load_verify_locations(certifi.where())
        return ctx

    def __init__(self, jid, password, recipient):
        self.g_jid = aioxmpp.JID.fromstr(jid)
        self.g_security_layer = aioxmpp.make_security_layer(password)._replace(
            ssl_context_factory=self.my_ssl_factory,
        )

        self.to_jid = aioxmpp.JID.fromstr(recipient)

        if os.name == 'nt':
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    def send(self, message):
        asyncio.run(self.async_send_messages([message]))

    def send_messages(self, messages):
        asyncio.run(self.async_send_messages(messages))

    async def async_send_messages(self, messages):
        client = aioxmpp.Client(
            self.g_jid,
            self.g_security_layer,
        )
        client.resumption_timeout = 0

        async with client.connected() as stream:
            for message_content in messages:
                msg = aioxmpp.Message(
                    to=self.to_jid,
                    type_=aioxmpp.MessageType.CHAT,
                )
                msg.body[None] = message_content

                await stream.send(msg)



import re

from yt_dlp.extractor.common import InfoExtractor

from yt_dlp.compat import (
    compat_urllib_parse,
    compat_urllib_parse_urlparse,
    compat_urllib_parse_unquote,
)

from yt_dlp.utils import (
    ExtractorError,
    int_or_none,
    url_or_none,
    urlencode_postdata,
    HEADRequest,
    determine_ext,
    mimetype2ext,
    encode_compat_str,
)


class OwncloudIE(InfoExtractor):
    IE_NAME = 'owncloud'

    _VALID_URL = r'''(?x)
            (?P<server>https?://(?:
                            .*\.?sciebo\.de|
                            cloud\.uni-koblenz-landau\.de
                        ))/s/
            (?P<id>[A-Za-z0-9\-_.]+)
            (?P<extra>/.*)?
        '''

    def _real_extract(self, url):

        mobj = re.match(self._VALID_URL, url)
        server = mobj.group('server')
        video_id = mobj.group('id')
        # url_extra = mobj.group('extra')

        landing_url = server + '/s/' + video_id
        landing_webpage, urlh = self._download_webpage_handle(url, landing_url, 'Downloading Owncloud landing page')
        opend_landing_url = urlh.geturl()

        password_protected = self._search_regex(
            r'<label[^>]+?for="(password)"', landing_webpage, 'password field', fatal=False, default=None
        )
        if password_protected is not None:
            # Password protected
            landing_webpage, urlh = self._verify_video_password(landing_webpage, opend_landing_url, video_id)

        landing_inputs = self._hidden_inputs(landing_webpage)

        title = landing_inputs.get('filename', 'Unknown title')
        # could be used for mimetype2ext
        # mimetype = landing_inputs.get('mimetype', None)

        filesize = landing_inputs.get('filesize', None)
        download_url = landing_inputs.get('downloadURL', None)

        if download_url is None:
            download_url = self._extend_to_download_url(urlh.geturl())

        ext_req = HEADRequest(download_url)
        ext_handle = self._request_webpage(ext_req, video_id, note='Determining extension')
        ext = self.urlhandle_detect_ext(ext_handle)

        formats = []
        formats.append(
            {
                'url': url_or_none(download_url),
                'ext': ext,
                'filesize': int_or_none(filesize),
            }
        )
        self._sort_formats(formats)

        return {'id': video_id, 'title': title, 'formats': formats}

    def urlhandle_detect_ext(self, url_handle):
        getheader = url_handle.headers.get

        def encode_compat_str_or_none(x, encoding='iso-8859-1', errors='ignore'):
            return encode_compat_str(x, encoding=encoding, errors=errors) if x else None

        cd = encode_compat_str_or_none(getheader('Content-Disposition'))
        if cd:
            m = re.match(
                r'''(?xi)
                attachment;\s*
                (?:filename\s*=[^;]+?;\s*)?                    # possible initial filename=...;, ignored
                filename(?P<x>\*)?\s*=\s*                      # filename/filename* =
                    (?(x)(?P<charset>\S+?)'[\w-]*'|(?P<q>")?)  # if * then charset'...' else maybe "
                    (?P<filename>(?(q)[^"]+(?=")|[^\s;]+))         # actual name of file
                ''',
                cd,
            )
            if m:
                m = m.groupdict()
                filename = m.get('filename')
                if m.get('x'):
                    try:
                        filename = compat_urllib_parse_unquote(filename, encoding=m.get('charset', 'utf-8'))
                    except LookupError:  # unrecognised character set name
                        pass
                e = determine_ext(filename, default_ext=None)
                if e:
                    return e

        ct = encode_compat_str_or_none(getheader('Content-Type'))
        return mimetype2ext(ct)

    def _extend_to_download_url(self, url: str) -> str:
        """
        Adds the string /download to a URL
        @param url: The URL where the string should be added.
        @return: The URL with the string.
        """

        url_parts = list(compat_urllib_parse_urlparse(url))
        url_parts[2] = url_parts[2].rstrip('/') + '/download'
        return compat_urllib_parse.urlunparse(url_parts)

    def _verify_video_password(self, webpage, url, video_id):
        password = self._downloader.params.get('videopassword')
        if password is None:
            raise ExtractorError(
                'This video is protected by a password, use the --video-password option', expected=True
            )
        requesttoken = self._search_regex(r'<input[^>]+?name="requesttoken" value="([^\"]+)"', webpage, 'requesttoken')
        data = urlencode_postdata({'requesttoken': requesttoken, 'password': password})

        validation_response, urlh = self._download_webpage_handle(
            url, video_id, note='Validating Password...', errnote='Wrong password?', data=data
        )

        password_protected = self._search_regex(
            r'<label[^>]+?for="(password)"', validation_response, 'password field', fatal=False, default=None
        )
        warning = self._search_regex(
            r'<div[^>]+?class="warning">([^<]*)</div>',
            validation_response,
            'warning',
            fatal=False,
            default="The password is wrong. Try again.",
        )
        if password_protected is not None:
            raise ExtractorError(f'Login failed, {self.IE_NAME} said: {warning!r}', expected=True)
        return validation_response, urlh



import re

from moodle_dl.moodle_connector.request_helper import RequestHelper, RequestRejectedError
from moodle_dl.state_recorder.course import Course
from moodle_dl.download_service.path_tools import PathTools
from moodle_dl.moodle_connector.moodle_constants import moodle_html_footer, moodle_html_header


class LessonsHandler:
    """
    Fetches and parses the various endpoints in Moodle for lesson entries.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_lessons(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Lessons List for all courses from the
        Moodle system
        @return: A Dictionary of all lessons,
                 indexed by courses, then lessons
        """
        # do this only if version is greater then 3.3
        # because mod_lesson_get_lessons_by_courses will fail
        if self.version < 2017051500:
            return {}

        print('\rDownloading lessons information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        lessons_result = self.request_helper.post_REST('mod_lesson_get_lessons_by_courses', extra_data)

        lessons = lessons_result.get('lessons', [])

        result = {}
        for lesson in lessons:
            # This is the instance id with which we can make the API queries.
            lesson_id = lesson.get('id', 0)
            lesson_name = lesson.get('name', 'unnamed lesson')
            lesson_intro = lesson.get('intro', '')
            lesson_course_module_id = lesson.get('coursemodule', 0)
            lesson_introfiles = lesson.get('introfiles', [])
            lesson_introfiles += lesson.get('mediafiles', [])
            course_id = lesson.get('course', 0)

            # normalize
            for lesson_file in lesson_introfiles:
                file_type = lesson_file.get('type', '')
                if file_type is None or file_type == '':
                    lesson_file.update({'type': 'lesson_introfile'})

            if lesson_intro != '':
                # Add Intro File
                intro_file = {
                    'filename': 'Lesson intro',
                    'filepath': '/',
                    'description': lesson_intro,
                    'type': 'description',
                }
                lesson_introfiles.append(intro_file)

            lesson_entry = {
                lesson_course_module_id: {
                    'id': lesson_id,
                    'name': lesson_name,
                    'intro': lesson_intro,
                    'files': lesson_introfiles,
                }
            }

            course_dic = result.get(course_id, {})

            course_dic.update(lesson_entry)

            result.update({course_id: course_dic})

        return result

    def fetch_lessons_files(self, userid: int, lessons: {}) -> {}:
        """
        Fetches for the lessons list of all courses the additionally
        entries. This is kind of waste of resources, because there
        is no API to get all entries at once.
        @param userid: the user id.
        @param lessons: the dictionary of lessons of all courses.
        @return: A Dictionary of all lessons,
                 indexed by courses, then lessons
        """
        # do this only if version is greater then 3.3
        # because mod_lesson_get_user_attempt will fail
        if self.version < 2017051500:
            return lessons

        counter = 0
        total = 0
        # count total lessons for nice console output
        for course_id in lessons:
            for lesson_id in lessons[course_id]:
                total += 1

        for course_id in lessons:
            for lesson_id in lessons[course_id]:
                counter += 1
                lesson = lessons[course_id][lesson_id]
                real_id = lesson.get('id', 0)
                data = {'lessonid': real_id, 'userid': userid, 'lessonattempt': 0}

                shorted_lesson_name = lesson.get('name', '')
                if len(shorted_lesson_name) > 17:
                    shorted_lesson_name = shorted_lesson_name[:15] + '..'

                print(
                    (
                        '\r'
                        + 'Downloading lesson infos'
                        + f' {counter:3d}/{total:3d}'
                        + f' [{shorted_lesson_name:<17}|{course_id:6}]\033[K'
                    ),
                    end='',
                )

                try:
                    attempt_result = self.request_helper.post_REST('mod_lesson_get_user_attempt', data)
                except RequestRejectedError:
                    continue

                lesson_files = self._get_files_of_attempt(attempt_result, lesson.get('name', ''))
                lesson['files'] += lesson_files

        return lessons

    def _get_files_of_attempt(self, attempt_result: {}, lesson_name: str) -> []:
        result = []

        answerpages = attempt_result.get('answerpages', [])
        # The review page should actually be generated here.
        # https://github.com/moodle/moodle/blob/511a87f5fc357f18a4c53911f6e6c7f7b526246e/mod/lesson/report.php#L278-L366

        # Grade is in: attempt_result.userstats.gradeinfo.earned  (max points: attempt_result.userstats.gradeinfo.total)
        # Take care, grade can be None

        grade = attempt_result.get('userstats', {}).get('gradeinfo', {}).get('earned', None)
        grade_total = attempt_result.get('userstats', {}).get('gradeinfo', {}).get('total', None)

        if grade is not None and grade_total is not None:
            grade_file = {
                'filename': 'grade',
                'filepath': '/',
                'timemodified': 0,
                'description': str(grade) + ' / ' + str(grade_total),
                'type': 'description',
            }
            result.append(grade_file)

        # build lesson HTML
        lesson_html = moodle_html_header
        attempt_filename = PathTools.to_valid_name(lesson_name)
        lesson_is_empty = True
        for counter, answerpage in enumerate(answerpages):
            page_id = answerpage.get('page', {}).get('id', 0)
            lesson_id = answerpage.get('page', {}).get('lessonid', 0)

            shorted_lesson_name = lesson_name
            if len(shorted_lesson_name) > 17:
                shorted_lesson_name = shorted_lesson_name[:15] + '..'

            print(
                (
                    '\r'
                    + 'Downloading lesson pages'
                    + f' {counter + 1:3d}/{len(answerpages):3d}'
                    + f' [{shorted_lesson_name:<17}|{lesson_id:6}]\033[K'
                ),
                end='',
            )

            data = {'lessonid': lesson_id, 'pageid': page_id, 'returncontents': 1}

            try:
                page_result = self.request_helper.post_REST('mod_lesson_get_page_data', data)
            except RequestRejectedError:
                continue

            pagecontent = page_result.get('pagecontent', '').split('<script>')[0]

            if pagecontent != '':
                lesson_is_empty = False

            lesson_html += pagecontent + '\n'

            page_files = page_result.get('contentfiles', [])
            for page_file in page_files:
                file_type = page_file.get('type', '')
                if file_type is None or file_type == '':
                    page_file.update({'type': 'lesson_file'})

            for page_file in page_files:
                new_page_file = True
                for attempt_file in result:
                    if re.sub(r"\/page_contents\/\d+\/", "/", attempt_file.get('fileurl', '')) == re.sub(
                        r"\/page_contents\/\d+\/", "/", page_file.get('fileurl', '')
                    ):
                        if (
                            attempt_file.get('filesize', 0) == page_file.get('filesize', 0)
                            # sometimes the teacher adds the same file for multiple answer pages with a
                            # different timestamp
                            # and attempt_file.get('timemodified', 0) == page_file.get('timemodified', 0)
                            and attempt_file.get('filename', '') == page_file.get('filename', '')
                        ):
                            new_page_file = False
                            break
                if new_page_file:
                    result.append(page_file)

        if not lesson_is_empty:
            lesson_html += moodle_html_footer
            attempt_file = {
                'filename': attempt_filename,
                'filepath': '/',
                'timemodified': 0,
                'html': lesson_html,
                'filter_urls_during_search_containing': ['/mod_lesson/page_contents/'],
                'type': 'html',
                'no_search_for_urls': True,
            }
            result.append(attempt_file)

        return result



# -*- coding: utf-8 -*-
import sys
import termios
import tty
import select


# idea from:
# https://repolinux.wordpress.com/2012/10/09/non-blocking-read-from-stdin-in-python/
# Thanks to REPOLINUX
def kbhit():
    return sys.stdin in select.select([sys.stdin], [], [], 0)[0]


# Initially taken from:
# http://code.activestate.com/recipes/134892/
# Thanks to Danny Yoo
def readchar(blocking=False):
    if not blocking and not kbhit():
        return None
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


def readkey():
    """Get a single character on Linux. If an escaped key is pressed, the
    following characters are read as well (see key_linux.py)."""

    c1 = readchar(blocking=True)

    if c1 == "\x03":
        raise KeyboardInterrupt

    if c1 != "\x1B":
        return c1

    c2 = readchar(blocking=True)
    if c2 not in "\x4F\x5B":
        return c1 + c2

    c3 = readchar(blocking=True)
    if c3 not in "\x31\x32\x33\x35\x36":
        return c1 + c2 + c3

    c4 = readchar(blocking=True)
    if c4 not in "\x30\x31\x33\x34\x35\x37\x38\x39":
        return c1 + c2 + c3 + c4

    c5 = readchar(blocking=True)
    return c1 + c2 + c3 + c4 + c5



import os
import html

from pathlib import Path
from yt_dlp.utils import sanitize_filename, remove_start


class PathTools:
    """A set of methods to create correct paths."""

    restricted_filenames = False

    @staticmethod
    def to_valid_name(name: str) -> str:
        """Filtering invalid characters in filenames and paths.

        Args:
            name (str): The string that will go through the filtering

        Returns:
            str: The filtered string, that can be used as a filename.
        """

        if name is None:
            return None

        # Moodle saves the title of a section in HTML-Format,
        # so we need to unescape the string
        name = html.unescape(name)

        name = name.replace('\n', ' ')
        name = name.replace('\r', ' ')
        name = name.replace('\t', ' ')
        name = name.replace('\xad', '')
        while '  ' in name:
            name = name.replace('  ', ' ')
        name = sanitize_filename(name, PathTools.restricted_filenames)
        name = name.strip('. ')
        name = name.strip()

        return name

    @staticmethod
    def sanitize_path(path: str):
        """
        @param path: A path to sanitize.
        @return: A path where every part was sanitized using to_valid_name.
        """
        drive_or_unc, _ = os.path.splitdrive(path)
        norm_path = os.path.normpath(remove_start(path, drive_or_unc)).split(os.path.sep)
        if drive_or_unc:
            norm_path.pop(0)

        sanitized_path = [
            path_part if path_part in ['.', '..'] else PathTools.to_valid_name(path_part) for path_part in norm_path
        ]

        if drive_or_unc:
            sanitized_path.insert(0, drive_or_unc + os.path.sep)
        return os.path.join(*sanitized_path)

    @staticmethod
    def path_of_file_in_module(
        storage_path: str, course_fullname: str, file_section_name: str, file_module_name: str, file_path: str
    ):
        """
        @param storage_path: The path where all files should be stored.
        @param course_fullname: The name of the course where the file is
                                located.
        @param file_section_name: The name of the section where the file
                                  is located.
        @param file_module_name: The name of the module where the file
                                 is located.
        @param file_path: The additional path of a file (subdirectory).
        @return: A path where the file should be saved.
        """
        path = str(
            Path(storage_path)
            / PathTools.to_valid_name(course_fullname)
            / PathTools.to_valid_name(file_section_name)
            / PathTools.to_valid_name(file_module_name)
            / PathTools.sanitize_path(file_path).strip('/')
        )
        return path

    @staticmethod
    def path_of_file(storage_path: str, course_fullname: str, file_section_name: str, file_path: str):
        """
        @param storage_path: The path where all files should be stored.
        @param course_fullname: The name of the course where the file is
                                located.
        @param file_section_name: The name of the section where the file
                                  is located.
        @param file_path: The additional path of a file (subdirectory).
        @return: A path where the file should be saved.
        """
        path = str(
            Path(storage_path)
            / PathTools.to_valid_name(course_fullname)
            / PathTools.to_valid_name(file_section_name)
            / PathTools.sanitize_path(file_path).strip('/')
        )
        return path

    @staticmethod
    def flat_path_of_file(storage_path: str, course_fullname: str, file_path: str):
        """
        @param storage_path: The path where all files should be stored.
        @param course_fullname: The name of the course where the file is
                                located.
        @param file_path: The additional path of a file (subdirectory).
        @return: A path where the file should be saved.
        """
        path = str(
            Path(storage_path)
            / PathTools.to_valid_name(course_fullname)
            / PathTools.sanitize_path(file_path).strip('/')
        )
        return path



import os
import ssl
import sys
import time
import shutil
import logging
import threading
from queue import Queue
import certifi

from yt_dlp.utils import format_bytes

from moodle_dl.state_recorder.course import Course, File
from moodle_dl.download_service.path_tools import PathTools
from moodle_dl.download_service.url_target import URLTarget
from moodle_dl.download_service.downloader import Downloader
from moodle_dl.moodle_connector.moodle_service import MoodleService
from moodle_dl.moodle_connector.ssl_helper import configure_ssl_context


class DownloadService:
    """
    DownloadService manages the queue of files to be downloaded and starts
    the Downloader threads which download all URLTargets.
    Furthermore DownloadService is responsible for logging live information
    and errors.
    """

    thread_count = 5

    def __init__(
        self,
        courses: [Course],
        moodle_service: MoodleService,
        storage_path: str,
        skip_cert_verify: bool = False,
        ignore_ytdl_errors: bool = False,
    ):
        """
        Initiates the DownloadService with all files that
        need to be downloaded. A URLTarget is created for each file.
        @param courses: A list of courses that contains all modified files.
        @param moodle_service: A reference to the moodle_service, currently
                               only to get to the state_recorder and the token.
        @param storage_path: The location where the files will be saved.
        """

        # How much threads should be created
        if 'pydevd' in sys.modules:
            # if debugging only one thread should be started
            DownloadService.thread_count = 1

        self.courses = courses
        self.state_recorder = moodle_service.recorder
        self.token = moodle_service.config_helper.get_token()
        self.storage_path = storage_path

        # The wait queue for all URL targets to be downloaded.
        self.queue = Queue(0)
        # A list of the created threads
        self.threads = []
        # A lock to stabilize thread insecure resources.
        # writing in DB
        self.db_lock = threading.Lock()
        # reading file system
        self.fs_lock = threading.Lock()

        # Sets the download options
        self.options = moodle_service.config_helper.get_download_options()
        # Add console parameters
        self.options.update({'ignore_ytdl_errors': ignore_ytdl_errors})

        # report is used to collect successful and failed downloads
        self.report = {'success': [], 'failure': []}
        # thread_report is used to get live reports from the threads
        self.thread_report = [
            {
                'total': 0,
                'percentage': 0,
                'old_extra_totalsize': None,
                'extra_totalsize': None,
                'current_url': '',
                'external_dl': None,
            }
            for i in range(self.thread_count)
        ]
        # Collects the total size of the files that needs to be downloaded.
        self.total_to_download = 0
        self.last_threads_total_downloaded = 0
        self.last_status_timestamp = time.time()
        self.total_files = 0

        # delete files, that should be deleted
        self.state_recorder.batch_delete_files(self.courses)

        if skip_cert_verify:
            self.ssl_context = ssl._create_unverified_context()
        else:
            self.ssl_context = ssl.create_default_context(cafile=certifi.where())
        configure_ssl_context(self.ssl_context)
        self.skip_cert_verify = skip_cert_verify

        # Prepopulate queue with any files that were given
        for course in self.courses:
            for file in course.files:
                if file.deleted is False:
                    self.total_to_download += file.content_filesize

                    save_destination = self.gen_path(self.storage_path, course, file)

                    self.queue.put(
                        URLTarget(
                            file,
                            course,
                            save_destination,
                            self.token,
                            self.thread_report,
                            self.fs_lock,
                            self.ssl_context,
                            self.skip_cert_verify,
                            self.options,
                        )
                    )

                    self.total_files += 1

        logging.debug('Queue contains %s URLTargets', self.total_files)

    @staticmethod
    def gen_path(storage_path: str, course: Course, file: File):
        """
        Generates the directory path where a file should be stored
        """
        course_name = course.fullname
        if course.overwrite_name_with is not None:
            course_name = course.overwrite_name_with

        # if a flat path is requested
        if not course.create_directory_structure:
            return PathTools.flat_path_of_file(storage_path, course_name, file.content_filepath)

        # If the file is located in a folder or in an assignment,
        # it should be saved in a sub-folder
        # (with the name of the module).
        if file.module_modname.endswith(('assign', 'folder', 'data', 'forum', 'quiz', 'lesson', 'workshop', 'page')):
            file_path = file.content_filepath
            if file.content_type == 'submission_file':
                file_path = os.path.join('/submissions/', file_path.strip('/'))

            return PathTools.path_of_file_in_module(
                storage_path, course_name, file.section_name, file.module_name, file_path
            )
        else:
            return PathTools.path_of_file(storage_path, course_name, file.section_name, file.content_filepath)

    def run(self):
        """
        Starts all threads to download the files and
        issues status messages at regular intervals.
        """
        if self.total_files == 0:
            return

        self._create_downloader_threads()

        print('\n' * (len(self.threads)), end='')
        old_status_message = ''
        while not self._downloader_complete():
            new_status_message = self._get_status_message()
            if old_status_message != new_status_message:
                print(new_status_message, end='')
                old_status_message = new_status_message
            time.sleep(0.5)

        self._clear_status_message()

    def _create_downloader_threads(self):
        """
        Creates all downloader threads, initiates them
        with the queue and starts them.
        """
        for i in range(self.thread_count):
            thread = Downloader(self.queue, self.report, self.state_recorder, i, self.db_lock)
            thread.start()
            self.threads.append(thread)

    def _downloader_complete(self) -> bool:
        """
        Checks if a thread is still running, if so then the downloaders
        are not finished yet.
        @return: status of the downloaders
        """
        finished_downloading = True
        for thread in self.threads:
            if thread.is_alive():
                finished_downloading = False
                break
        return finished_downloading

    def _get_status_message(self) -> str:
        """
        Creates a string that combines the status messages of all threads.
        The current download progress of a file is displayed in percent
        per Thread.
        A total display is also created, showing the total amount downloaded
        in relation to what still needs to be downloaded.
        @return: A status message string
        """

        # to limit the output to one line
        limits = shutil.get_terminal_size()

        # Starting with a carriage return to overwrite the last message
        progressmessage = f'\033[{len(self.threads)}A\r'

        threads_status_message = ''
        threads_total_downloaded = 0
        for thread in self.threads:

            i = thread.thread_id
            # A thread status contains it id and the progress
            # of the current file
            thread_percentage = self.thread_report[i]['percentage']
            thread_current_url = self.thread_report[i]['current_url']
            if self.thread_report[i]['external_dl'] is not None:
                thread_current_url = 'ExtDL: ' + self.thread_report[i]['external_dl']

            if not thread.is_alive():
                thread_percentage = 100
                thread_current_url = 'Finished!'

            if len(thread_current_url) + 13 > limits.columns:
                thread_current_url = thread_current_url[0 : limits.columns - 15] + '..'

            threads_status_message += f'\033[KT{int(i):2}: {int(thread_percentage):3}% - {thread_current_url}\n'

            threads_total_downloaded += self.thread_report[i]['total']

            extra_totalsize = self.thread_report[i]['extra_totalsize']
            if extra_totalsize is not None and extra_totalsize != -1:
                self.total_to_download += extra_totalsize
                self.thread_report[i]['extra_totalsize'] = -1

        progressmessage += threads_status_message

        percentage = 100
        if self.total_to_download != 0:
            percentage = int(threads_total_downloaded * 100 / self.total_to_download)

        # The overall progress also includes the total size that needs to be
        # downloaded and the size that has already been downloaded.
        progressmessage_line = (
            f'Total: {percentage:3}%'
            + f' {format_bytes(threads_total_downloaded):>12} / {format_bytes(self.total_to_download):<12}'
        )

        progressmessage_line += f" | Files: {len(self.report['success']):>5} / {self.total_files:<5}"

        diff_to_last_status = threads_total_downloaded - self.last_threads_total_downloaded

        speed = self.calc_speed(self.last_status_timestamp, time.time(), diff_to_last_status)
        progressmessage_line += ' | ' + self.format_speed(speed)

        if len(progressmessage_line) > limits.columns:
            progressmessage_line = progressmessage_line[0 : limits.columns]
        progressmessage_line = '\033[K' + progressmessage_line

        progressmessage += progressmessage_line

        self.last_status_timestamp = time.time()
        self.last_threads_total_downloaded = threads_total_downloaded

        return progressmessage

    @staticmethod
    def calc_speed(start, now, byte_count):
        dif = now - start
        if byte_count <= 0 or dif < 0.001:  # One millisecond
            return None
        return float(byte_count) / dif

    @staticmethod
    def format_speed(speed):
        if speed is None:
            return f"{'---b/s':10}"
        speed_text = format_bytes(speed) + '/s'
        return f'{speed_text:10}'

    def _clear_status_message(self):
        print(f'\033[{len(self.threads)}A\r', end='')

        print('\033[K\n' * (len(self.threads)), end='')
        print('\033[K', end='')

        print(f'\033[{len(self.threads)}A\r', end='')

    def get_failed_url_targets(self):
        """
        Return a list of failed Downloads, as a list of URLTargets.
        """
        return self.report['failure']



import sys
import shutil
import logging

from pathlib import Path
from getpass import getpass
from urllib.parse import urlparse
from distutils.version import StrictVersion

from yt_dlp.utils import determine_ext

from moodle_dl.utils import cutie
from moodle_dl.utils.logger import Log
from moodle_dl.config_service.config_helper import ConfigHelper
from moodle_dl.state_recorder.course import Course
from moodle_dl.state_recorder.state_recorder import StateRecorder
from moodle_dl.moodle_connector import login_helper
from moodle_dl.moodle_connector import sso_token_receiver
from moodle_dl.moodle_connector.cookie_handler import CookieHandler
from moodle_dl.moodle_connector.results_handler import ResultsHandler
from moodle_dl.moodle_connector.pages_handler import PagesHandler
from moodle_dl.moodle_connector.forums_handler import ForumsHandler
from moodle_dl.moodle_connector.folders_handler import FoldersHandler
from moodle_dl.moodle_connector.quizzes_handler import QuizzesHandler
from moodle_dl.moodle_connector.lessons_handler import LessonsHandler
from moodle_dl.moodle_connector.workshops_handler import WorkshopsHandler
from moodle_dl.moodle_connector.databases_handler import DatabasesHandler
from moodle_dl.moodle_connector.assignments_handler import AssignmentsHandler
from moodle_dl.moodle_connector.first_contact_handler import FirstContactHandler
from moodle_dl.moodle_connector.request_helper import RequestRejectedError, RequestHelper


class MoodleService:
    def __init__(
        self,
        config_helper: ConfigHelper,
        storage_path: str,
        skip_cert_verify: bool = False,
        log_responses: bool = False,
    ):
        self.config_helper = config_helper
        self.storage_path = storage_path
        self.recorder = StateRecorder(Path(storage_path) / 'moodle_state.db')
        self.skip_cert_verify = skip_cert_verify

        self.log_responses_to = None
        if log_responses:
            self.log_responses_to = str(Path(storage_path) / 'responses.log')

    def interactively_acquire_token(
        self, use_stored_url: bool = False, username: str = None, password: str = None
    ) -> str:
        """
        Walks the user through executing a login into the Moodle-System to get
        the Token and saves it.
        @return: The Token for Moodle.
        """

        automated = False
        stop_automatic_generation = False
        if username is not None and password is not None:
            automated = True

        if not automated:
            print('[The following Credentials are not saved, it is only used temporarily to generate a login token.]')

        moodle_token = None
        while moodle_token is None:
            if stop_automatic_generation and automated:
                break

            if not use_stored_url:
                moodle_url = input('URL of Moodle:   ')

                use_http = False
                if moodle_url.startswith('http://'):
                    Log.error(
                        'Warning: You have entered an insecure URL! Are you sure that the Moodle is'
                        + ' not accessible via `https://`? All your data will be transferred'
                        + ' insecurely! If your Moodle is accessible via `https://`, then run'
                        + ' the process again using `https://` to protect your data.'
                    )
                    use_http = True
                elif not moodle_url.startswith('https://'):
                    Log.error('The url of your moodle must start with `https://`')
                    continue

                moodle_uri = urlparse(moodle_url)

                moodle_domain, moodle_path = self._split_moodle_uri(moodle_uri)

            else:
                moodle_domain = self.config_helper.get_moodle_domain()
                moodle_path = self.config_helper.get_moodle_path()
                use_http = self.config_helper.get_use_http()

            if username is not None:
                moodle_username = username
                stop_automatic_generation = True
            else:
                moodle_username = input('Username for Moodle:   ')

            if password is not None:
                moodle_password = password
            else:
                moodle_password = getpass('Password for Moodle [no output]:   ')

            try:
                moodle_token, moodle_privatetoken = login_helper.obtain_login_token(
                    moodle_username,
                    moodle_password,
                    moodle_domain,
                    moodle_path,
                    self.skip_cert_verify,
                    use_http,
                )

            except RequestRejectedError as error:
                Log.error(f'Login Failed! ({error}) Please try again.')
            except (ValueError, RuntimeError) as error:
                Log.error(f'Error while communicating with the Moodle System! ({error}) Please try again.')
            except ConnectionError as error:
                Log.error(str(error))

        if automated is True and moodle_token is None:
            sys.exit(1)

        # Saves the created token and the successful Moodle parameters.
        self.config_helper.set_property('token', moodle_token)
        if moodle_privatetoken is not None:
            self.config_helper.set_property('privatetoken', moodle_privatetoken)
        self.config_helper.set_property('moodle_domain', moodle_domain)
        self.config_helper.set_property('moodle_path', moodle_path)
        if use_http is True:
            self.config_helper.set_property('use_http', use_http)

        return moodle_token

    def interactively_acquire_sso_token(self, use_stored_url: bool = False) -> str:
        """
        Walks the user through the receiving of a SSO token for the
        Moodle-System and saves it.
        @return: The Token for Moodle.
        """
        if not use_stored_url:

            moodle_url = input('URL of Moodle:   ')

            moodle_uri = urlparse(moodle_url)

            moodle_domain, moodle_path = self._split_moodle_uri(moodle_uri)

        else:
            moodle_domain = self.config_helper.get_moodle_domain()
            moodle_path = self.config_helper.get_moodle_path()

        use_http = self.config_helper.get_use_http()
        scheme = 'https://'
        if use_http:
            scheme = 'http://'

        atomatic_procedure_warning = (
            'Between version 3.8.1 and 3.8.2 a change was added to'
            + ' Moodle so that automatic copying of the SSO token'
            + ' might not work.'
        )
        allow_automatic = True
        try:
            version = RequestHelper(
                moodle_domain,
                moodle_path,
                skip_cert_verify=self.skip_cert_verify,
                use_http=use_http,
            ).get_simple_moodle_version()

            if StrictVersion(version) > StrictVersion("3.8.1") and StrictVersion(version) <= StrictVersion("3.8.2"):
                Log.warning(atomatic_procedure_warning + '\nYou can still try it, your version is: ' + str(version))
            if StrictVersion(version) > StrictVersion("3.8.2"):
                Log.warning(
                    'For your Moodle version there is currently no automatic way to copy the SSO token implemented,'
                    + f' your version is: {str(version)}. You will have to proceed manually.'
                )
                allow_automatic = False

        except (ConnectionError, RequestRejectedError):
            Log.warning(
                atomatic_procedure_warning
                + '\nThe version of your Moodle could not be detected, you can still try the automatic procedure.'
            )

        do_automatic = False
        if allow_automatic:
            print(' If you want to copy the login-token manual, you will be guided through the manual copy process.')
            do_automatic = cutie.prompt_yes_or_no('Do you want to try to receive the SSO token automatically?')

        print('Please log into Moodle on this computer and then visit the following address in your web browser: ')

        if do_automatic:
            print(
                scheme
                + moodle_domain
                + moodle_path
                + 'admin/tool/mobile/launch.php?service='
                + 'moodle_mobile_app&passport=12345&'
                + 'urlscheme=http%3A%2F%2Flocalhost'
            )
            moodle_token = sso_token_receiver.receive_token()
        else:
            print(
                scheme
                + moodle_domain
                + moodle_path
                + 'admin/tool/mobile/launch.php?service='
                + 'moodle_mobile_app&passport=12345&urlscheme=moodledownloader'
            )

            print(
                'If you open the link in the browser, no web page should'
                + ' load, instead an error will occur. Open the'
                + ' developer console (press F12) and go to the Network Tab,'
                + ' if there is no error, reload the web page.'
            )

            print(
                'Copy the link address of the website that could not be'
                + ' loaded (right click, then click on Copy, then click'
                + ' on copy link address).'
            )

            print(
                'The script expects a URL that looks something like this:'
                + '`moodledownloader://token=$apptoken`.'
                + ' Where $apptoken looks random. In reality it is a Base64'
                + ' encoded hash and the token we need to access moodle.'
            )

            token_address = input('Then insert the address here:   ')

            moodle_token, moodle_privatetoken = sso_token_receiver.extract_token(token_address)
            if moodle_token is None:
                raise ValueError('Invalid URL!')

        # Saves the created token and the successful Moodle parameters.
        self.config_helper.set_property('token', moodle_token)
        if moodle_privatetoken is not None:
            self.config_helper.set_property('privatetoken', moodle_privatetoken)
        self.config_helper.set_property('moodle_domain', moodle_domain)
        self.config_helper.set_property('moodle_path', moodle_path)

        return moodle_token

    def fetch_state(self) -> [Course]:
        """
        Gets the current status of the configured Moodle account and compares
        it with the last known status for changes. It does not change the
        known state, nor does it download the files.
        @return: List with detected changes
        """
        logging.debug('Fetching current Moodle State...')

        token = self.config_helper.get_token()
        privatetoken = self.config_helper.get_privatetoken()
        moodle_domain = self.config_helper.get_moodle_domain()
        moodle_path = self.config_helper.get_moodle_path()
        use_http = self.config_helper.get_use_http()

        request_helper = RequestHelper(
            moodle_domain,
            moodle_path,
            token,
            self.skip_cert_verify,
            self.log_responses_to,
            use_http,
        )
        first_contact_handler = FirstContactHandler(request_helper)
        results_handler = ResultsHandler(request_helper, moodle_domain, moodle_path)

        download_course_ids = self.config_helper.get_download_course_ids()
        download_public_course_ids = self.config_helper.get_download_public_course_ids()
        dont_download_course_ids = self.config_helper.get_dont_download_course_ids()
        download_submissions = self.config_helper.get_download_submissions()
        download_databases = self.config_helper.get_download_databases()
        download_forums = self.config_helper.get_download_forums()
        download_quizzes = self.config_helper.get_download_quizzes()
        download_lessons = self.config_helper.get_download_lessons()
        download_workshops = self.config_helper.get_download_workshops()
        download_also_with_cookie = self.config_helper.get_download_also_with_cookie()

        courses = []
        filtered_courses = []
        cookie_handler = None

        print('\rDownloading account information\033[K', end='')

        userid, version = self.config_helper.get_userid_and_version()
        if userid is None or version is None:
            userid, version = first_contact_handler.fetch_userid_and_version()
        else:
            first_contact_handler.version = version
        assignments_handler = AssignmentsHandler(request_helper, version)
        databases_handler = DatabasesHandler(request_helper, version)
        forums_handler = ForumsHandler(request_helper, version)
        quizzes_handler = QuizzesHandler(request_helper, version)
        lessons_handler = LessonsHandler(request_helper, version)
        workshops_handler = WorkshopsHandler(request_helper, version)
        pages_handler = PagesHandler(request_helper, version)
        folders_handler = FoldersHandler(request_helper, version)

        results_handler.setVersion(version)

        if download_also_with_cookie:
            # generate a new cookie if necessary
            cookie_handler = CookieHandler(request_helper, version, self.storage_path)
            cookie_handler.check_and_fetch_cookies(privatetoken, userid)

        courses_list = first_contact_handler.fetch_courses(userid)
        courses = []
        # Filter unselected courses
        for course in courses_list:
            if ResultsHandler.should_download_course(course.id, download_course_ids, dont_download_course_ids):
                courses.append(course)

        public_courses_list = first_contact_handler.fetch_courses_info(download_public_course_ids)
        for course in public_courses_list:
            courses.append(course)

        assignments = assignments_handler.fetch_assignments(courses)
        if download_submissions:
            assignments = assignments_handler.fetch_submissions(userid, assignments)

        databases = databases_handler.fetch_databases(courses)
        if download_databases:
            databases = databases_handler.fetch_database_files(databases)

        forums = forums_handler.fetch_forums(courses)
        if download_forums:
            last_timestamps_per_forum = self.recorder.get_last_timestamps_per_forum()
            forums = forums_handler.fetch_forums_posts(forums, last_timestamps_per_forum)

        quizzes = quizzes_handler.fetch_quizzes(courses)
        if download_quizzes:
            quizzes = quizzes_handler.fetch_quizzes_files(userid, quizzes)

        lessons = lessons_handler.fetch_lessons(courses)
        if download_lessons:
            lessons = lessons_handler.fetch_lessons_files(userid, lessons)

        workshops = workshops_handler.fetch_workshops(courses)
        if download_workshops:
            workshops = workshops_handler.fetch_workshops_files(userid, workshops)

        pages = pages_handler.fetch_pages(courses)

        folders = folders_handler.fetch_folders(courses)

        courses = self.add_options_to_courses(courses)
        index = 0
        for course in courses:
            index += 1

            # to limit the output to one line
            limits = shutil.get_terminal_size()

            shorted_course_name = course.fullname
            if len(course.fullname) > 17:
                shorted_course_name = course.fullname[:15] + '..'

            into = '\rDownloading course information'

            status_message = into + f' {index:3d}/{len(courses):3d} [{shorted_course_name:<17}|{course.id:6}]'

            if len(status_message) > limits.columns:
                status_message = status_message[0 : limits.columns]

            print(status_message + '\033[K', end='')

            course_fetch_addons = {
                'assign': assignments.get(course.id, {}),
                'data': databases.get(course.id, {}),
                'forum': forums.get(course.id, {}),
                'quiz': quizzes.get(course.id, {}),
                'lesson': lessons.get(course.id, {}),
                'workshop': workshops.get(course.id, {}),
                'page': pages.get(course.id, {}),
                'folder': folders.get(course.id, {}),
            }
            results_handler.set_fetch_addons(course_fetch_addons)
            course.files = results_handler.fetch_files(course)

            filtered_courses.append(course)
        print('')

        logging.debug('Checking for changes...')
        changes = self.recorder.changes_of_new_version(filtered_courses)

        # Filter changes
        changes = self.add_options_to_courses(changes)
        changes = self.filter_courses(changes, self.config_helper, cookie_handler, courses_list + public_courses_list)

        return changes

    def add_options_to_courses(self, courses: [Course]):
        """
        Updates a array of courses with its options
        """
        options_of_courses = self.config_helper.get_options_of_courses()
        for course in courses:
            options = options_of_courses.get(str(course.id), None)
            if options is not None:
                course.overwrite_name_with = options.get('overwrite_name_with', None)
                course.create_directory_structure = options.get('create_directory_structure', True)
                course.excluded_sections = options.get("excluded_sections", [])

        return courses

    @staticmethod
    def filter_courses(
        changes: [Course],
        config_helper: ConfigHelper,
        cookie_handler: CookieHandler = None,
        courses_list: [Course] = None,
    ) -> [Course]:
        """
        Filters the changes course list from courses that
        should not get downloaded
        @param config_helper: ConfigHelper to obtain all the diffrent filter configs
        @param cookie_handler: CookieHandler to check if the cookie is valid
        @param courses_list: A list of all courses that are available online
        @return: filtered changes course list
        """

        download_course_ids = config_helper.get_download_course_ids()
        download_public_course_ids = config_helper.get_download_public_course_ids()
        dont_download_course_ids = config_helper.get_dont_download_course_ids()
        download_submissions = config_helper.get_download_submissions()
        download_descriptions = config_helper.get_download_descriptions()
        download_links_in_descriptions = config_helper.get_download_links_in_descriptions()
        download_databases = config_helper.get_download_databases()
        download_quizzes = config_helper.get_download_quizzes()
        download_lessons = config_helper.get_download_lessons()
        download_workshops = config_helper.get_download_workshops()
        exclude_file_extensions = config_helper.get_exclude_file_extensions()
        download_also_with_cookie = config_helper.get_download_also_with_cookie()
        if cookie_handler is not None:
            download_also_with_cookie = cookie_handler.test_cookies()

        filtered_changes = []

        for course in changes:
            if not ResultsHandler.should_download_course(
                course.id, download_course_ids + download_public_course_ids, dont_download_course_ids
            ):
                # Filter courses that should not be downloaded
                continue

            if courses_list is not None:
                not_online = True
                # Filter courses that are not available online
                for online_course in courses_list:
                    if online_course.id == course.id:
                        not_online = False
                        break
                if not_online:
                    Log.warning(f'The Moodle course with id {course.id} is no longer available online.')
                    logging.warning('The Moodle course with id %d is no longer available online.', course.id)
                    continue

            course_files = []
            for file in course.files:
                # Filter Files based on options
                if (
                    # Filter Assignment Submission Files
                    (download_submissions or (not (file.module_modname.endswith('assign') and file.deleted)))
                    # Filter Description Files (except the forum posts)
                    and (
                        download_descriptions
                        or file.content_type != 'description'
                        or (
                            file.module_modname == 'forum'
                            and file.content_type == 'description'
                            and file.content_filename != 'Forum intro'
                        )
                    )
                    # Filter Database Files
                    and (download_databases or file.content_type != 'database_file')
                    # Filter Quiz Files
                    and (download_quizzes or (not (file.module_modname.endswith('quiz') and file.deleted)))
                    # Filter Lesson Files
                    and (download_lessons or (not (file.module_modname.endswith('lesson') and file.deleted)))
                    # Filter Workshops Files
                    and (download_workshops or (not (file.module_modname.endswith('workshop') and file.deleted)))
                    # Filter Files that requiere a Cookie
                    and (download_also_with_cookie or (not file.module_modname.startswith('cookie_mod-')))
                    # Exclude files whose file extension is blacklisted
                    and (determine_ext(file.content_filename) not in exclude_file_extensions)
                    # Exclude files that are in excluded sections
                    and (ResultsHandler.should_download_section(file.section_id, course.excluded_sections))
                ):
                    course_files.append(file)
            course.files = course_files

            # Filter Description URLs
            course_files = []
            for file in course.files:
                if not file.content_type == 'description-url':
                    course_files.append(file)

                elif download_links_in_descriptions:
                    add_description_url = True
                    for test_file in course.files:
                        if file.content_fileurl == test_file.content_fileurl:
                            if test_file.content_type != 'description-url':
                                # If a URL in a description also exists as a real link in the course,
                                # then ignore this URL
                                add_description_url = False
                                break
                            elif file.module_id > test_file.module_id:
                                # Always use the link from the older description.
                                add_description_url = False
                                break

                    if add_description_url:
                        course_files.append(file)
            course.files = course_files

            if len(course.files) > 0:
                filtered_changes.append(course)

        return filtered_changes

    @staticmethod
    def _split_moodle_uri(moodle_uri: str):
        """
        Splits a given Moodle-Uri into the domain and the installation path
        @return: moodle_domain, moodle_path as strings
        """

        moodle_domain = moodle_uri.netloc
        moodle_path = moodle_uri.path
        if not moodle_path.endswith('/'):
            moodle_path = moodle_path + '/'

        if moodle_path == '':
            moodle_path = '/'

        return moodle_domain, moodle_path



#! /usr/bin/env python3

"""
Command-line User Tools for simpler Input

Source: https://github.com/Kamik423/cutie/blob/master/cutie.py

Authors: Hans and Kamik423
License: MIT (https://github.com/Kamik423/cutie/blob/master/license.md)
"""

import os
import shutil
import getpass

from typing import List, Optional
from colorama import init

import moodle_dl.utils.readchar as readchar

terminal = os.getenv('TERM')
if terminal is None:
    init()
else:
    init(convert=False, strip=False)


class DefaultKeys:
    """List of default keybindings.
    Attributes:
        interrupt(List[str]): Keys that cause a keyboard interrupt.
        select(List[str]): Keys that trigger list element selection.
        select_all(List[str]): Keys that trigger selection of all list elements.
        confirm(List[str]): Keys that trigger list confirmation.
        delete(List[str]): Keys that trigger character deletion.
        down(List[str]): Keys that select the element below.
        up(List[str]): Keys that select the element above.
    """

    interrupt: List[str] = [readchar.key.CTRL_C, readchar.key.CTRL_D]
    select: List[str] = [readchar.key.SPACE]
    select_all: List[str] = [readchar.key.CTRL_A]
    confirm: List[str] = [readchar.key.ENTER, readchar.key.CR, readchar.key.LF]
    delete: List[str] = [readchar.key.BACKSPACE]
    down: List[str] = [readchar.key.DOWN, 'j']
    up: List[str] = [readchar.key.UP, 'k']
    tab: List[str] = ['\t']


def get_number(
    prompt: str, min_value: Optional[float] = None, max_value: Optional[float] = None, allow_float: bool = True
) -> float:
    """Get a number from user input.
    If an invalid number is entered the user will be prompted again.
    Args:
        prompt (str): The prompt asking the user to input.
        min_value (float, optional): The [inclusive] minimum value.
        max_value (float, optional): The [inclusive] maximum value.
        allow_float (bool, optional): Allow floats or force integers.
    Returns:
        float: The number input by the user.
    """
    return_value: Optional[float] = None
    while return_value is None:
        input_value = input(prompt + ' ')
        try:
            return_value = float(input_value)
        except ValueError:
            print('Not a valid number.\033[K\033[1A\r\033[K', end='')
        if not allow_float and return_value is not None:
            if return_value != int(return_value):
                print('Has to be an integer.\033[K\033[1A\r\033[K', end='')
                return_value = None
        if min_value is not None and return_value is not None:
            if return_value < min_value:
                print(f'Has to be at least {min_value}.\033[K\033[1A\r\033[K', end='')
                return_value = None
        if max_value is not None and return_value is not None:
            if return_value > max_value:
                print(f'Has to be at most {max_value}.\033[1A\r\033[K', end='')
                return_value = None
        if return_value is not None:
            break
    print('\033[K', end='')
    if allow_float:
        return return_value
    return int(return_value)


def secure_input(prompt: str) -> str:
    """Get secure input without showing it in the command line.
    Args:
        prompt (str): The prompt asking the user to input.
    Returns:
        str: The secure input.
    """
    return getpass.getpass(prompt + ' ')


def select(
    options: List[str],
    caption_indices: Optional[List[int]] = None,
    deselected_prefix: str = '\033[1m[ ]\033[0m ',
    selected_prefix: str = '\033[1m[\033[32;1mx\033[0;1m]\033[0m ',
    caption_prefix: str = '',
    selected_index: int = 0,
    confirm_on_select: bool = True,
    reserved_lines: int = 3,
) -> int:
    """Select an option from a list.
    Args:
        options (List[str]): The options to select from.
        caption_indices (List[int], optional): Non-selectable indices.
        deselected_prefix (str, optional): Prefix for deselected option ([ ]).
        selected_prefix (str, optional): Prefix for selected option ([x]).
        caption_prefix (str, optional): Prefix for captions ().
        selected_index (int, optional): The index to be selected at first.
        confirm_on_select (bool, optional): Select keys also confirm.
        reserved_lines (int, optional): How many lines of the terminal are
            reserved and should not be used.
    Returns:
        int: The index that has been selected.
    """
    if caption_indices is None:
        caption_indices = []

    max_index = len(options)
    max_lines = len(options) + 1  # Last line is for empty line

    # Lines that were output in the previous interation
    lines_printed = 0
    # By how many entries is the list shifted
    shift = 0

    while True:
        print(f'\033[{lines_printed}A')

        console_lines = shutil.get_terminal_size().lines
        # Extra empty line for correct terminal behavior
        view_hight = max(0, min(console_lines - reserved_lines, max_lines))
        # View consists of
        # (top-indicator)
        # entries
        # (bottom-indicator)
        # empty line

        #  Darstellbaren Einträge =  view_hight - top-indicator
        if shift > (max_lines - 1) - (view_hight - 2):
            shift = (max_lines - 1) - (view_hight - 2)

        # Darstellbaren Einträge normal =  view_hight - top-indicator - bottom-indicator - empty line
        data_bottom = shift + (view_hight - 3)
        if shift == 0:
            # we do not need to print "x more lines above...", so we have one more entry line
            data_bottom += 1
        else:
            print(f'\033[K{shift} more lines above...')

        if data_bottom + 1 >= len(options):
            data_bottom = len(options)

        for i in range(shift, data_bottom):
            option = options[i]
            console_columns = shutil.get_terminal_size().columns - 5
            printable_option = option.expandtabs().replace('\n', ' ').replace('\r', ' ')
            if len(printable_option) > console_columns:
                printable_option = printable_option[: (console_columns - 2)] + '..'

            if i not in caption_indices:
                print(f'\033[K{selected_prefix if i == selected_index else deselected_prefix}{printable_option}')
            elif i in caption_indices:
                print(f'\033[K{caption_prefix}{printable_option}')

        if data_bottom != len(options):
            print(f'{len(options) - data_bottom} more lines below...\033[K')

        lines_printed = view_hight

        keypress = readchar.readkey()
        if keypress in DefaultKeys.up:
            new_index = selected_index
            while new_index > 0:
                new_index -= 1
                if new_index not in caption_indices:
                    selected_index = new_index
                    if selected_index < shift:
                        if shift == 2:
                            shift = 0
                        else:
                            shift = selected_index
                    break
        elif keypress in DefaultKeys.down:
            new_index = selected_index
            while new_index < max_index - 1:
                new_index += 1
                if new_index not in caption_indices:
                    selected_index = new_index
                    if selected_index >= data_bottom and data_bottom != max_index:
                        shift = selected_index - (view_hight - 4)
                    break
        elif keypress in DefaultKeys.confirm or confirm_on_select and keypress in DefaultKeys.select:
            break
        elif keypress in DefaultKeys.interrupt:
            raise KeyboardInterrupt
    return selected_index


def select_multiple(
    options: List[str],
    caption_indices: Optional[List[int]] = None,
    deselected_unticked_prefix: str = '\033[1m( )\033[0m ',
    deselected_ticked_prefix: str = '\033[1m(\033[32mx\033[0;1m)\033[0m ',
    selected_unticked_prefix: str = '\033[32;1m{ }\033[0m ',
    selected_ticked_prefix: str = '\033[32;1m{x}\033[0m ',
    caption_prefix: str = '',
    ticked_indices: Optional[List[int]] = None,
    cursor_index: int = 0,
    minimal_count: int = 0,
    maximal_count: Optional[int] = None,
    hide_confirm: bool = False,
    deselected_confirm_label: str = '\033[1m(( confirm ))\033[0m',
    selected_confirm_label: str = '\033[1;32m{{ confirm }}\033[0m',
    reserved_lines: int = 3,
) -> List[int]:
    """Select multiple options from a list.
    Args:
        options (List[str]): The options to select from.
        caption_indices (List[int], optional): Non-selectable indices.
        deselected_unticked_prefix (str, optional): Prefix for lines that are
            not selected and not ticked (( )).
        deselected_ticked_prefix (str, optional): Prefix for lines that are
            not selected but ticked ((x)).
        selected_unticked_prefix (str, optional): Prefix for lines that are
            selected but not ticked ({ }).
        selected_ticked_prefix (str, optional): Prefix for lines that are
            selected and ticked ({x}).
        caption_prefix (str, optional): Prefix for captions ().
        ticked_indices (List[int], optional): Indices that are
            ticked initially.
        cursor_index (int, optional): The index the cursor starts at.
        minimal_count (int, optional): The minimal amount of lines
            that have to be ticked.
        maximal_count (int, optional): The maximal amount of lines
            that have to be ticked.
        hide_confirm (bool, optional): Hide the confirm button.
            This causes <ENTER> to confirm the entire selection and not just
            tick the line.
        deselected_confirm_label (str, optional): The confirm label
            if not selected ((( confirm ))).
        selected_confirm_label (str, optional): The confirm label
            if selected ({{ confirm }}).
        reserved_lines (int, optional): How many lines of the terminal are
            reserved and should not be used.
    Returns:
        List[int]: The indices that have been selected
    """
    if caption_indices is None:
        caption_indices = []
    if ticked_indices is None:
        ticked_indices = []
    max_index = len(options) - (1 if hide_confirm else 0)
    max_lines = len(options) + 2  # Last two line are for confirm / error / bottom-indicator + empty line
    error_message = ''

    # Lines that were output in the previous interation
    lines_printed = 0
    # By how many entries is the list shifted
    shift = 0

    while True:
        print(f'\033[{lines_printed}A')

        console_lines = shutil.get_terminal_size().lines
        # Extra empty line for correct terminal behavior
        view_hight = max(0, min(console_lines - reserved_lines, max_lines))
        # View consists of
        # (top-indicator)
        # entries
        # (bottom-indicator / error / confirm)
        # empty line

        #  Darstellbaren Einträge =  view_hight - bottom-indicator/confirm/error - top-indicator
        if shift > (max_lines - 1) - (view_hight - 2):
            shift = (max_lines - 1) - (view_hight - 2)

        # Darstellbaren Einträge normal =  view_hight - top-indicator - bottom-indicator/confirm/error - empty line
        data_bottom = shift + (view_hight - 3)
        if shift == 0:
            # we do not need to print "x more lines above...", so we have one more entry line
            data_bottom += 1
        else:
            print(f'\033[K{shift} more lines above...')

        if data_bottom > len(options):
            data_bottom = len(options)

        for i in range(shift, data_bottom):
            option = options[i]
            console_columns = shutil.get_terminal_size().columns - 5
            printable_option = option.expandtabs().replace('\n', ' ').replace('\r', ' ')
            if len(printable_option) > console_columns:
                printable_option = printable_option[: (console_columns - 2)] + '..'

            prefix = ''
            if i in caption_indices:
                prefix = caption_prefix
            elif i == cursor_index:
                if i in ticked_indices:
                    prefix = selected_ticked_prefix
                else:
                    prefix = selected_unticked_prefix
            else:
                if i in ticked_indices:
                    prefix = deselected_ticked_prefix
                else:
                    prefix = deselected_unticked_prefix
            print(f'\033[K{prefix}{printable_option}')

        if data_bottom == len(options):
            # we do not need to print "x more lines below...", instead we print the confirm label or an error
            if hide_confirm:
                print(f'{error_message}\033[K')
            else:
                if cursor_index == max_index:
                    print(f'{selected_confirm_label} {error_message}\033[K')
                else:
                    print(f'{deselected_confirm_label} {error_message}\033[K')
        else:
            print(f'{len(options) - data_bottom} more lines below... {error_message}\033[K')

        lines_printed = view_hight

        error_message = ''
        keypress = readchar.readkey()
        if keypress in DefaultKeys.up:
            new_index = cursor_index
            while new_index > 0:
                new_index -= 1
                if new_index not in caption_indices:
                    cursor_index = new_index
                    if cursor_index < shift:
                        if shift == 2:
                            shift = 0
                        else:
                            shift = cursor_index
                    break
        elif keypress in DefaultKeys.down:
            new_index = cursor_index
            while new_index + 1 <= max_index:
                new_index += 1
                if new_index not in caption_indices:
                    cursor_index = new_index
                    if cursor_index >= data_bottom and data_bottom != len(options):
                        shift = cursor_index - (view_hight - 4)
                    break
        elif keypress in DefaultKeys.select:
            if cursor_index in ticked_indices:
                if len(ticked_indices) - 1 >= minimal_count:
                    ticked_indices.remove(cursor_index)
            elif maximal_count is not None:
                if len(ticked_indices) + 1 <= maximal_count:
                    ticked_indices.append(cursor_index)
            else:
                ticked_indices.append(cursor_index)
        elif keypress in DefaultKeys.confirm:
            if minimal_count > len(ticked_indices):
                error_message = f'Must select at least {minimal_count} options'
            elif maximal_count is not None and maximal_count < len(ticked_indices):
                error_message = f'Must select at most {maximal_count} options'
            else:
                break
        elif keypress in DefaultKeys.select_all:
            for i in range(0, len(options)):
                if i not in ticked_indices:
                    ticked_indices.append(i)
        elif keypress in DefaultKeys.interrupt:
            raise KeyboardInterrupt
    print('\033[1A\033[K', end='', flush=True)
    return ticked_indices


def prompt_yes_or_no(
    question: str,
    yes_text: str = 'Yes',
    no_text: str = 'No',
    has_to_match_case: bool = False,
    enter_empty_confirms: bool = True,
    default_is_yes: bool = False,
    deselected_prefix: str = '  ',
    selected_prefix: str = '\033[31m>\033[0m ',
    char_prompt: bool = True,
) -> Optional[bool]:
    """Prompt the user to input yes or no.
    Args:
        question (str): The prompt asking the user to input.
        yes_text (str, optional): The text corresponding to 'yes'.
        no_text (str, optional): The text corresponding to 'no'.
        has_to_match_case (bool, optional): Does the case have to match.
        enter_empty_confirms (bool, optional): Does enter on empty string work.
        default_is_yes (bool, optional): Is yes selected by default (no).
        deselected_prefix (str, optional): Prefix if something is deselected.
        selected_prefix (str, optional): Prefix if something is selected (> )
        char_prompt (bool, optional): Add a [Y/N] to the prompt.
    Returns:
        Optional[bool]: The bool what has been selected.
    """
    is_yes = default_is_yes
    is_selected = enter_empty_confirms
    current_message = ''
    if char_prompt:
        if default_is_yes:
            yn_prompt = f' ({yes_text[0]}/{no_text[0].lower()}) '
        else:
            yn_prompt = f' ({yes_text[0].lower()}/{no_text[0]}) '
    else:
        yn_prompt = ': '
    print()
    while True:
        yes = is_yes and is_selected
        no = not is_yes and is_selected
        print('\033[K' f'{selected_prefix if yes else deselected_prefix}{yes_text}')
        print('\033[K' f'{selected_prefix if no else deselected_prefix}{no_text}')
        print('\033[3A\r\033[K' f'{question}{yn_prompt}{current_message}', end='', flush=True)
        keypress = readchar.readkey()
        if keypress in DefaultKeys.down or keypress in DefaultKeys.up:
            is_yes = not is_yes
            is_selected = True
            current_message = yes_text if is_yes else no_text
        elif keypress in DefaultKeys.interrupt:
            raise KeyboardInterrupt
        elif keypress in DefaultKeys.confirm:
            if is_selected:
                break
        elif keypress is not None:
            if keypress in DefaultKeys.tab:
                pass
            elif keypress in DefaultKeys.delete:
                if current_message:
                    current_message = current_message[:-1]
            else:
                current_message += keypress

            match_yes = yes_text
            match_no = no_text
            match_text = current_message
            if not has_to_match_case:
                match_yes = match_yes.upper()
                match_no = match_no.upper()
                match_text = match_text.upper()
            if match_no.startswith(match_text):
                is_selected = True
                is_yes = False
            elif match_yes.startswith(match_text):
                is_selected = True
                is_yes = True
            else:
                is_selected = False

            if keypress in DefaultKeys.tab and is_selected:
                current_message = yes_text if is_yes else no_text
        print()
    print('\033[K\n\033[K\n\033[K\n\033[3A')
    return is_selected and is_yes



from moodle_dl.moodle_connector.request_helper import RequestHelper, RequestRejectedError
from moodle_dl.state_recorder.course import Course


class WorkshopsHandler:
    """
    Fetches and parses the various endpoints in Moodle for workshop entries.
    """

    def __init__(self, request_helper: RequestHelper, version: int):
        self.request_helper = request_helper
        self.version = version

    def fetch_workshops(self, courses: [Course]) -> {int: {int: {}}}:
        """
        Fetches the Workshops List for all courses from the
        Moodle system
        @return: A Dictionary of all workshops,
                 indexed by courses, then workshops
        """
        # do this only if version is greater then 3.4
        # because mod_workshop_get_workshops_by_courses will fail
        if self.version < 2017111300:
            return {}

        print('\rDownloading workshops information\033[K', end='')

        # We create a dictionary with all the courses we want to request.
        extra_data = {}
        courseids = {}
        for index, course in enumerate(courses):
            courseids.update({str(index): course.id})

        extra_data.update({'courseids': courseids})

        workshops_result = self.request_helper.post_REST('mod_workshop_get_workshops_by_courses', extra_data)

        workshops = workshops_result.get('workshops', [])

        result = {}
        for workshop in workshops:
            # This is the instance id with which we can make the API queries.
            workshop_id = workshop.get('id', 0)
            workshop_name = workshop.get('name', 'unnamed workshop')
            workshop_course_module_id = workshop.get('coursemodule', 0)
            workshop_introfiles = workshop.get('introfiles', [])
            workshop_introfiles += workshop.get('instructauthorsfiles', [])
            workshop_introfiles += workshop.get('instructreviewersfiles', [])
            workshop_introfiles += workshop.get('conclusionfiles', [])
            course_id = workshop.get('course', 0)

            # normalize
            for workshop_file in workshop_introfiles:
                file_type = workshop_file.get('type', '')
                if file_type is None or file_type == '':
                    workshop_file.update({'type': 'workshop_introfile'})

            workshop_intro = workshop.get('intro', '')
            if workshop_intro != '':
                # Add Intro File
                intro_file = {
                    'filename': 'Workshop intro',
                    'filepath': '/',
                    'description': workshop_intro,
                    'type': 'description',
                }
                workshop_introfiles.append(intro_file)

            workshop_instructauthors = workshop.get('instructauthors', '')
            if workshop_instructauthors != '':
                # Add Instructions for submission File
                instructauthors_file = {
                    'filename': 'Instructions for submission',
                    'filepath': '/',
                    'description': workshop_instructauthors,
                    'type': 'description',
                }
                workshop_introfiles.append(instructauthors_file)

            workshop_instructreviewers = workshop.get('instructreviewers', '')
            if workshop_instructreviewers != '':
                # Add Instructions for assessment File
                instructreviewers_file = {
                    'filename': 'Instructions for assessment',
                    'filepath': '/',
                    'description': workshop_instructreviewers,
                    'type': 'description',
                }
                workshop_introfiles.append(instructreviewers_file)

            workshop_conclusion = workshop.get('conclusion', '')
            if workshop_conclusion != '':
                # Add Conclusion File
                conclusion_file = {
                    'filename': 'Conclusion',
                    'filepath': '/',
                    'description': workshop_conclusion,
                    'type': 'description',
                }
                workshop_introfiles.append(conclusion_file)

            workshop_entry = {
                workshop_course_module_id: {
                    'id': workshop_id,
                    'name': workshop_name,
                    'intro': workshop_intro,
                    'files': workshop_introfiles,
                }
            }

            course_dic = result.get(course_id, {})

            course_dic.update(workshop_entry)

            result.update({course_id: course_dic})

        return result

    def fetch_workshops_files(self, userid: int, workshops: {}) -> {}:
        """
        Fetches for the workshops list of all courses the additionally
        entries. This is kind of waste of resources, because there
        is no API to get all entries at once.
        @param userid: the user id.
        @param workshops: the dictionary of workshops of all courses.
        @return: A Dictionary of all workshops,
                 indexed by courses, then workshops
        """
        # do this only if version is greater then 3.3
        # because mod_quiz_get_user_attempts will fail
        if self.version < 2017111300:
            return workshops

        counter = 0
        total = 0
        # count total workshops for nice console output
        for course_id in workshops:
            for workshop_id in workshops[course_id]:
                total += 1

        for course_id in workshops:
            for workshop_id in workshops[course_id]:
                counter += 1
                workshop = workshops[course_id][workshop_id]
                real_id = workshop.get('id', 0)
                data = {'workshopid': real_id, 'userid': userid}

                shorted_workshop_name = workshop.get('name', '')
                if len(shorted_workshop_name) > 17:
                    shorted_workshop_name = shorted_workshop_name[:15] + '..'

                print(
                    (
                        '\r'
                        + 'Downloading workshop infos'
                        + f' {counter:3d}/{total:3d}'
                        + f' [{shorted_workshop_name:<17}|{course_id:6}]\033[K'
                    ),
                    end='',
                )

                try:
                    submissions_result = self.request_helper.post_REST('mod_workshop_get_submissions', data)
                except RequestRejectedError:
                    continue

                try:
                    reviewer_assessments_result = self.request_helper.post_REST(
                        'mod_workshop_get_reviewer_assessments', data
                    )
                except RequestRejectedError:
                    reviewer_assessments_result = {}

                try:
                    grades_result = self.request_helper.post_REST('mod_workshop_get_grades', data)
                except RequestRejectedError:
                    grades_result = {}

                workshop_files = self._get_files_of_workshop(
                    submissions_result, reviewer_assessments_result, grades_result
                )
                workshop['files'] += workshop_files

        return workshops

    def _get_files_of_workshop(self, submissions_result: {}, reviewer_assessments_result: {}, grades_result: {}) -> []:
        result = []

        submissions = submissions_result.get('submissions', [])

        reviewer_assessments = reviewer_assessments_result.get('assessments', [])

        for reviewer_assessment in reviewer_assessments:
            # reviewer_assessment_id = reviewer_assessment.get('id', 0)
            reviewer_assessment_submissionid = reviewer_assessment.get('submissionid', 0)
            # reviewer_assessment_reviewerid = reviewer_assessment.get('reviewerid', 0)

            # Get attached files
            reviewer_assessment_files = reviewer_assessment.get('feedbackcontentfiles', [])
            reviewer_assessment_files += reviewer_assessment.get('feedbackattachmentfiles', [])

            feedbackauthor = reviewer_assessment.get('feedbackauthor', '')
            if feedbackauthor != '':
                # Add Feedback for the author
                feedbackauthor_file = {
                    'filename': 'Feedback for the author',
                    'filepath': '/',
                    'description': feedbackauthor,
                    'type': 'description',
                }
                reviewer_assessment_files.append(feedbackauthor_file)

            feedbackreviewer = reviewer_assessment.get('feedbackreviewer', '')
            if feedbackreviewer != '':
                # Add Feedback for the reviewer
                feedbackreviewer_file = {
                    'filename': 'Feedback for the reviewer',
                    'filepath': '/',
                    'description': feedbackreviewer,
                    'type': 'description',
                }
                reviewer_assessment_files.append(feedbackreviewer_file)

            # Get submissions of assessments
            data = {'submissionid': reviewer_assessment_submissionid}
            try:
                submission_result = self.request_helper.post_REST('mod_workshop_get_submission', data)
            except RequestRejectedError:
                submission_result = None

            if submission_result is not None:
                submission = submission_result.get('submission', {})
                submission['files'] = reviewer_assessment_files
                submissions.append(submission)

        assessmentlongstrgrade = grades_result.get('assessmentlongstrgrade', '')
        if assessmentlongstrgrade != '':
            # Add Assessment Grade
            assessmentlongstrgrade_file = {
                'filename': 'Assessment grade',
                'filepath': '/',
                'description': assessmentlongstrgrade,
                'type': 'description',
            }
            result.append(assessmentlongstrgrade_file)

        submissionlongstrgrade = grades_result.get('submissionlongstrgrade', '')
        if submissionlongstrgrade != '':
            # Add Submission Grade
            submissionlongstrgrade_file = {
                'filename': 'Submission grade',
                'filepath': '/',
                'description': submissionlongstrgrade,
                'type': 'description',
            }
            result.append(submissionlongstrgrade_file)

        for submission in submissions:
            submission_id = submission.get('id', 0)
            submission_timemodified = submission.get('timemodified', 0)
            submission_title = submission.get('title', 0)
            submission_content = submission.get('content', 0)

            filepath = f'/submissions {submission_id}/'

            submission_files = submission.get('contentfiles', [])
            submission_files += submission.get('attachmentfiles', [])
            submission_files += submission.get('files', [])

            for submission_file in submission_files:
                file_type = submission_file.get('type', '')
                if file_type is None or file_type == '':
                    submission_file.update({'type': 'workshop_file'})
                submission_file.update({'filepath': filepath})

            if submission_content != '':
                # Add submission content file
                content_file = {
                    'filename': submission_title,
                    'filepath': filepath,
                    'description': submission_content,
                    'timemodified': submission_timemodified,
                    'type': 'description',
                }
                submission_files.append(content_file)
            result += submission_files

        return result



import sys
import shutil

from pathlib import Path

from moodle_dl.utils import cutie
from moodle_dl.utils.logger import Log
from moodle_dl.state_recorder.course import Course
from moodle_dl.config_service.config_helper import ConfigHelper
from moodle_dl.moodle_connector.results_handler import ResultsHandler
from moodle_dl.moodle_connector.first_contact_handler import FirstContactHandler
from moodle_dl.moodle_connector.request_helper import RequestRejectedError, RequestHelper


class ConfigService:
    def __init__(self, config_helper: ConfigHelper, storage_path: str, skip_cert_verify: bool = False):
        self.config_helper = config_helper
        self.storage_path = storage_path
        self.skip_cert_verify = skip_cert_verify

        token = self.config_helper.get_token()
        moodle_domain = self.config_helper.get_moodle_domain()
        moodle_path = self.config_helper.get_moodle_path()
        use_http = self.config_helper.get_use_http()

        request_helper = RequestHelper(moodle_domain, moodle_path, token, self.skip_cert_verify, use_http=use_http)
        self.first_contact_handler = FirstContactHandler(request_helper)

    def interactively_acquire_config(self):
        """
        Guides the user through the process of configuring the downloader
        for the courses to be downloaded and in what way
        """
        courses = []
        try:
            userid, version = self.config_helper.get_userid_and_version()
            if userid is None or version is None:
                userid, version = self.first_contact_handler.fetch_userid_and_version()
                self._select_should_userid_and_version_be_saved(userid, version)
            else:
                self.first_contact_handler.version = version

            courses = self.first_contact_handler.fetch_courses(userid)

        except (RequestRejectedError, ValueError, RuntimeError, ConnectionError) as error:
            Log.error(f'Error while communicating with the Moodle System! ({error})')
            sys.exit(1)

        self._select_courses_to_download(courses)
        self._set_options_of_courses(courses)
        self._select_should_download_submissions()
        self._select_should_download_descriptions()
        self._select_should_download_links_in_descriptions()
        self._select_should_download_databases()
        self._select_should_download_forums()
        self._select_should_download_quizzes()
        self._select_should_download_lessons()
        self._select_should_download_workshops()
        self._select_should_download_linked_files()
        self._select_should_download_also_with_cookie()

    def interactively_add_all_visible_courses(self):
        """
        Guides the user through the process of adding all visible courses
        to the list of courses to download in the configuration
        """
        print('')
        Log.info(
            'It is possible to automatically complete the moodle-dl configuration'
            + ' with all the courses you can see on your moodle. These are either'
            + ' courses to which you have the appropriate rights to see the'
            + ' course or the course is visible without enrollment.'
        )

        Log.critical(
            'This process can take several minutes for large Moodels, as is common at'
            + ' large universities. Timeout is set to 20 minutes.'
        )

        print('')

        add_all_visible_courses = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to add all visible courses of your Moodle to the configuration?'),
            default_is_yes=False,
        )

        if not add_all_visible_courses:
            return
        else:
            Log.warning(
                'Please wait for the result, this may take several minutes.'
                + ' In addition to adding the courses to the configuration,'
                + ' it will also create an `all_courses.json` file with all'
                + ' the courses available on your Moodle.'
            )

        courses = []
        all_visible_courses = []
        try:
            userid, version = self.config_helper.get_userid_and_version()
            if userid is None or version is None:
                userid, version = self.first_contact_handler.fetch_userid_and_version()
            else:
                self.first_contact_handler.version = version

            courses = self.first_contact_handler.fetch_courses(userid)
            log_all_courses_to = str(Path(self.storage_path) / 'all_courses.json')
            all_visible_courses = self.first_contact_handler.fetch_all_visible_courses(log_all_courses_to)

        except (RequestRejectedError, ValueError, RuntimeError, ConnectionError) as error:
            Log.error(f'Error while communicating with the Moodle System! ({error})')
            sys.exit(1)

        # Filter out courses the user is enroled in
        filtered_all_courses = []
        for visible_course in all_visible_courses:
            add_to_final_list = True
            for course in courses:
                if visible_course.id == course.id:
                    add_to_final_list = False
                    break
            if add_to_final_list:
                filtered_all_courses.append(visible_course)

        # Update Public Courses IDs
        download_public_course_ids = self.config_helper.get_download_public_course_ids()
        # Update Course settings List for all new Courses
        options_of_courses = self.config_helper.get_options_of_courses()
        for course in filtered_all_courses:
            current_course_settings = options_of_courses.get(str(course.id), None)

            # create default settings
            if current_course_settings is None:
                current_course_settings = {
                    'original_name': course.fullname,
                    'overwrite_name_with': None,
                    'create_directory_structure': True,
                }

                options_of_courses.update({str(course.id): current_course_settings})

            if course.id not in download_public_course_ids:
                download_public_course_ids.append(course.id)

        self.config_helper.set_property('options_of_courses', options_of_courses)
        self.config_helper.set_property('download_public_course_ids', download_public_course_ids)

    def _select_should_userid_and_version_be_saved(self, userid, version):
        """
        Asks the user if the userid and version should be saved in the configuration
        """

        print('')
        Log.info(
            'The user id and version number of Moodle are downloaded'
            + ' at the beginning of each run of the downloader.'
            + ' Since this data rarely changes, it can be saved in the'
            + ' configuration.'
        )

        Log.critical(f'Your user id is `{userid}` and the moodle version is `{version}`')

        print('')

        save_userid_and_version = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to store the user id and version number of Moodle in the configuration?'),
            default_is_yes=False,
        )

        if save_userid_and_version:
            Log.warning(
                'Remember to delete the version number from the'
                + ' configuration once Moodle has been updated'
                + ' and then run the configurator again!'
            )

            self.config_helper.set_property('userid', userid)
            self.config_helper.set_property('version', version)

        self.section_seperator()

    def _select_courses_to_download(self, courses: [Course]):
        """
        Asks the user for the courses that should be downloaded.
        @param courses: All available courses
        """
        download_course_ids = self.config_helper.get_download_course_ids()
        dont_download_course_ids = self.config_helper.get_dont_download_course_ids()

        print('')
        Log.info(
            'To avoid downloading all Moodle courses you are enrolled in, you can select which courses you want'
            + ' to download here. You can either create a whitelist or a blacklist for this purpose.'
            + '\n\n- With a whitelist only the courses you have selected will be downloaded,'
            + ' any courses you enroll in in the future will not be downloaded automatically unless'
            + ' you add them to the list.'
            + '\n- With a blacklist all selected courses will not be downloaded, all other courses that are not'
            + ' on the list will be downloaded, but if you enroll in a new course online in the future that is not'
            + ' on your blacklist it will automatically be downloaded as well. '
        )
        print('')
        use_whitelist = len(dont_download_course_ids) == 0

        use_whitelist = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to create a whitelist or blacklist for your courses?'),
            default_is_yes=use_whitelist,
            yes_text='Whitelist',
            no_text='Blacklist',
        )

        choices = []
        defaults = []
        for i, course in enumerate(courses):
            choices.append(f'{int(course.id):5}\t{course.fullname}')

            should_download = ResultsHandler.should_download_course(
                course.id, download_course_ids, dont_download_course_ids
            )
            if should_download and use_whitelist:
                defaults.append(i)
            elif not should_download and not use_whitelist:
                defaults.append(i)

        if use_whitelist:
            Log.special('Which of the courses should be downloaded?')
        else:
            Log.special('Which of the courses should NOT be downloaded?')
        Log.info('[You can select with the space bar and confirm your selection with the enter key]')
        print('')
        selected_courses = cutie.select_multiple(options=choices, ticked_indices=defaults)

        course_ids = []
        for i, course in enumerate(courses):
            if i in selected_courses:
                course_ids.append(course.id)

        if use_whitelist:
            self.config_helper.set_property('download_course_ids', course_ids)
            self.config_helper.remove_property('dont_download_course_ids')
        elif not use_whitelist:
            self.config_helper.set_property('dont_download_course_ids', course_ids)
            self.config_helper.remove_property('download_course_ids')

    def _select_sections_to_download(self, sections: [{}], excluded: [int]) -> [int]:
        """
        Asks the user for the sections that should be downloaded.
        @param sections: All available sections
        @param excluded sections currently excluded
        """

        choices = []
        defaults = []
        for i, section in enumerate(sections):
            section_id = section.get("id")
            choices.append(f"{int(section_id):5}\t{section.get('name')}")

            if ResultsHandler.should_download_section(section_id, excluded):
                defaults.append(i)

        Log.special('Which of the sections should be downloaded?')
        Log.info('[You can select with the space bar and confirm your selection with the enter key]')
        print('')
        selected_sections = cutie.select_multiple(options=choices, ticked_indices=defaults)

        dont_download_section_ids = []
        for i, section in enumerate(sections):
            if i not in selected_sections:
                dont_download_section_ids.append(section.get("id"))

        return dont_download_section_ids

    def _set_options_of_courses(self, courses: [Course]):
        """
        Let the user set special options for every single course
        """
        download_course_ids = self.config_helper.get_download_course_ids()
        dont_download_course_ids = self.config_helper.get_dont_download_course_ids()

        self.section_seperator()
        Log.info(
            'You can set special settings for every single course.\n'
            + 'You can set these options:\n'
            + ' - A different name for the course\n'
            + ' - If a directory structure should be created for the course'
            + ' [create_directory_structure (cfs)]\n'
            + ' - Which of the sections should be downloaded (default all).'
        )
        print('')

        while True:

            choices = []
            choices_courses = []

            options_of_courses = self.config_helper.get_options_of_courses()

            choices.append('None')

            for course in courses:
                if ResultsHandler.should_download_course(course.id, download_course_ids, dont_download_course_ids):

                    current_course_settings = options_of_courses.get(str(course.id), None)

                    # create default settings
                    if current_course_settings is None:
                        current_course_settings = {
                            'original_name': course.fullname,
                            'overwrite_name_with': None,
                            'create_directory_structure': True,
                            'excluded_sections': [],
                        }

                    # create list of options
                    overwrite_name_with = current_course_settings.get('overwrite_name_with', None)

                    create_directory_structure = current_course_settings.get('create_directory_structure', True)

                    if overwrite_name_with is not None and overwrite_name_with != course.fullname:
                        choices.append(
                            (
                                f'{int(course.id):5}\t{overwrite_name_with} ({course.fullname})'
                                + f' cfs={create_directory_structure}'
                            )
                        )

                    else:
                        choices.append(f'{int(course.id):5}\t{course.fullname}  cfs={create_directory_structure}')

                    choices_courses.append(course)

            print('')
            Log.special('For which of the following course do you want to change the settings?')
            print('[Confirm your selection with the Enter key]')
            print('')

            selected_course = cutie.select(options=choices)
            if selected_course == 0:
                break
            else:
                sel = choices_courses[selected_course - 1]
                self._change_settings_of(sel, options_of_courses)

    def _change_settings_of(self, course: Course, options_of_courses: {}):
        """
        Ask for a new Name for the course.
        Then asks if a file structure should be created.
        """

        current_course_settings = options_of_courses.get(str(course.id), None)

        # create default settings
        if current_course_settings is None:
            current_course_settings = {
                'original_name': course.fullname,
                'overwrite_name_with': None,
                'create_directory_structure': True,
                'excluded_sections': [],
            }

        changed = False

        # Ask for new name
        overwrite_name_with = input(f'Enter a new name for this Course [leave blank for "{course.fullname}"]:   ')

        if overwrite_name_with == '':
            overwrite_name_with = None

        if (
            overwrite_name_with != course.fullname
            and current_course_settings.get('overwrite_name_with', None) != overwrite_name_with
        ):
            current_course_settings.update({'overwrite_name_with': overwrite_name_with})
            changed = True

        # Ask if a file structure should be created
        create_directory_structure = current_course_settings.get('create_directory_structure', True)

        create_directory_structure = cutie.prompt_yes_or_no(
            Log.special_str('Should a directory structure be created for this course?'),
            default_is_yes=create_directory_structure,
        )

        if create_directory_structure is not current_course_settings.get('create_directory_structure', True):
            changed = True
            current_course_settings.update({'create_directory_structure': create_directory_structure})

        excluded_sections = current_course_settings.get('excluded_sections', [])

        change_excluded_sections_promt = 'Do you want to exclude individual sections of this course from download?'
        if len(excluded_sections) > 0:
            change_excluded_sections_promt = (
                'Do you want to change the selection of sections that should not be'
                + f' downloaded? Currently {len(excluded_sections)} sections of this course are excluded from download.'
            )

        change_excluded_sections = cutie.prompt_yes_or_no(
            Log.special_str(change_excluded_sections_promt),
            default_is_yes=(len(excluded_sections) > 0),
        )

        if change_excluded_sections:
            Log.info('Please wait for a moment, the information about the course sections gets downloaded.')
            sections = self.first_contact_handler.fetch_sections(course.id)

            dont_download_section_ids = self._select_sections_to_download(sections, excluded_sections)

            if dont_download_section_ids != current_course_settings.get('excluded_sections', []):
                changed = True
                current_course_settings.update({'excluded_sections': dont_download_section_ids})

        if changed:
            options_of_courses.update({str(course.id): current_course_settings})
            self.config_helper.set_property('options_of_courses', options_of_courses)

    def _select_should_download_submissions(self):
        """
        Asks the user if submissions should be downloaded
        """
        download_submissions = self.config_helper.get_download_submissions()

        self.section_seperator()
        Log.info(
            'Submissions are files that you or a teacher have uploaded'
            + ' to your assignments. Moodle does not provide an'
            + ' interface for downloading information from all'
            + ' submissions to a course at once.'
        )
        Log.warning('Therefore, it may be slow to monitor changes to submissions.')
        print('')

        download_submissions = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to download submissions of your assignments?'),
            default_is_yes=download_submissions,
        )

        self.config_helper.set_property('download_submissions', download_submissions)

    def _select_should_download_databases(self):
        """
        Asks the user if databases should be downloaded
        """
        download_databases = self.config_helper.get_download_databases()

        self.section_seperator()
        Log.info(
            'In the database module of Moodle data can be stored'
            + ' structured with information. Often it is also'
            + ' possible for students to upload data there.  Because'
            + ' the implementation of the downloader has not yet been'
            + ' optimized at this point, it is optional to download the'
            + ' databases. Currently only files are downloaded, thumbails'
            + ' are ignored.'
        )
        print('')

        download_databases = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to download databases of your courses?'), default_is_yes=download_databases
        )

        self.config_helper.set_property('download_databases', download_databases)

    def _select_should_download_forums(self):
        """
        Asks the user if forums should be downloaded
        """
        download_forums = self.config_helper.get_download_forums()

        self.section_seperator()
        Log.info('In forums, students and teachers can discuss and exchange information together.')
        print('')

        download_forums = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to download forums of your courses?'), default_is_yes=download_forums
        )

        self.config_helper.set_property('download_forums', download_forums)

    def _select_should_download_quizzes(self):
        """
        Asks the user if quizzes should be downloaded
        """
        download_quizzes = self.config_helper.get_download_quizzes()

        self.section_seperator()
        Log.info(
            'Quizzes are tests that a student must complete in a course and are graded on.'
            + ' Only quizzes that are in progress or have been completed will be downloaded.'
        )
        print('')

        download_quizzes = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to download quizzes of your courses?'), default_is_yes=download_quizzes
        )

        self.config_helper.set_property('download_quizzes', download_quizzes)

    def _select_should_download_lessons(self):
        """
        Asks the user if lessons should be downloaded
        """
        download_lessons = self.config_helper.get_download_lessons()

        self.section_seperator()
        Log.info(
            'Lessons are a kind of self-teaching with pages of information and other pages with questions to answer.'
            + ' A student can be graded on their answers after completing a lesson. Currently, only lessons without'
            + ' the answers are downloaded. The answers are potentially also available for download,'
            + ' but this has not been implemented.'
        )
        print('')

        download_lessons = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to download lessons of your courses?'), default_is_yes=download_lessons
        )

        self.config_helper.set_property('download_lessons', download_lessons)

    def _select_should_download_workshops(self):
        """
        Asks the user if workshops should be downloaded
        """
        download_workshops = self.config_helper.get_download_workshops()

        self.section_seperator()
        Log.info(
            'Workshops function according to the peer review process.'
            + ' Students can make submissions and have to assess submissions of other students. '
        )
        print('')

        download_workshops = cutie.prompt_yes_or_no(
            Log.special_str('Do you want to download workshops of your courses?'), default_is_yes=download_workshops
        )

        self.config_helper.set_property('download_workshops', download_workshops)

    def _select_should_download_descriptions(self):
        """
        Asks the user if descriptions should be downloaded
        """
        download_descriptions = self.config_helper.get_download_descriptions()

        self.section_seperator()
        Log.info(
            'In Moodle courses, descriptions can be added to all kinds'
            + ' of resources, such as files, tasks, assignments or simply'
            + ' free text. These descriptions are usually unnecessary to'
            + ' download because you have already read the information or'
            + ' know it from context. However, there are situations where'
            + ' it might be interesting to download these descriptions. The'
            + ' descriptions are created as Markdown files and can be'
            + ' deleted as desired.'
        )
        Log.debug(
            'Creating the description files does not take extra time, but they can be annoying'
            + ' if they only contain unnecessary information.'
        )

        print('')

        download_descriptions = cutie.prompt_yes_or_no(
            Log.special_str('Would you like to download descriptions of the courses you have selected?'),
            default_is_yes=download_descriptions,
        )

        self.config_helper.set_property('download_descriptions', download_descriptions)

    def _select_should_download_links_in_descriptions(self):
        """
        Asks the user if links in descriptions should be downloaded
        """
        download_links_in_descriptions = self.config_helper.get_download_links_in_descriptions()

        self.section_seperator()
        Log.info(
            'In the descriptions of files, sections, assignments or courses the teacher can add links to webpages,'
            + ' files or videos. That links can point to a internal page on moodle or to an external webpage.'
        )
        print('')

        download_links_in_descriptions = cutie.prompt_yes_or_no(
            Log.special_str('Would you like to download links in descriptions?'),
            default_is_yes=download_links_in_descriptions,
        )

        self.config_helper.set_property('download_links_in_descriptions', download_links_in_descriptions)

    def _select_should_download_linked_files(self):
        """
        Asks the user if linked files should be downloaded
        """
        download_linked_files = self.config_helper.get_download_linked_files()

        self.section_seperator()
        Log.info(
            'In Moodle courses the teacher can also link to external'
            + ' files. This can be audio, video, text or anything else.'
            + ' In particular, the teacher can link to Youtube videos.'
        )
        Log.debug('To download videos correctly you have to install ffmpeg. ')

        Log.error('These files can increase the download volume considerably.')

        Log.info(
            'If you want to filter the external links by their domain,'
            + ' you can manually set a whitelist and a blacklist'
            + ' (https://github.com/C0D3D3V/Moodle-Downloader-2/'
            + 'wiki/Download-(external)-linked-files'
            + ' for more details).'
        )
        Log.warning(
            'Please note that the size of the external files is determined during the download, so the total size'
            + ' changes during the download.'
        )
        print('')

        download_linked_files = cutie.prompt_yes_or_no(
            Log.special_str('Would you like to download linked files of the courses you have selected?'),
            default_is_yes=download_linked_files,
        )

        self.config_helper.set_property('download_linked_files', download_linked_files)

    def _select_should_download_also_with_cookie(self):
        """
        Ask the user whether files for which a cookie is required should be downloaded.
        """
        download_also_with_cookie = self.config_helper.get_download_also_with_cookie()

        self.section_seperator()
        Log.info(
            'Descriptions may contain links to files that require a browser cookie so they can be downloaded.'
            + ' There are also several Moodle plugins that cannot be displayed in the Moodle app,'
            + ' so you need a browser cookie to download these plugin files.'
        )

        Log.debug(
            'The Moodle browser cookie is created using your private token and stored in the `Configs.txt` file.'
            + ' As long as this option is activated, the file always contains a valid cookie.'
        )

        if self.config_helper.get_privatetoken() is None:
            Log.error(
                'Currently no private token is stored in the configuration.'
                + ' Create a private token with moodle-dl --new-token (if necessary with --sso)'
            )

        print('')

        download_also_with_cookie = cutie.prompt_yes_or_no(
            Log.special_str('Would you like to download files for which a cookie is required?'),
            default_is_yes=download_also_with_cookie,
        )

        self.config_helper.set_property('download_also_with_cookie', download_also_with_cookie)

    def section_seperator(self):
        """Print a seperator line."""
        print('\n' + '-' * shutil.get_terminal_size().columns + '\n')
